@@ -23,12 +23,13 @@
 using namespace std;
 
 #include "osd/OSD.h"
+#include "os/FileStore.h"
+#include "mon/MonClient.h"
 #include "include/ceph_features.h"
 
 #include "common/config.h"
 
 #include "mon/MonMap.h"
-#include "mon/MonClient.h"
 
 
 #include "msg/Messenger.h"
@@ -45,6 +46,8 @@ using namespace std;
 
 #include "perfglue/heap_profiler.h"
 
+#include "include/assert.h"
+
 #define dout_subsys ceph_subsys_osd
 
 OSD *osd = NULL;
@@ -185,18 +188,26 @@ int main(int argc, const char **argv)
       derr << "Unable to get a Ceph keyring." << dendl;
       return 1;
     }
+
     EntityName ename(g_conf->name);
     EntityAuth eauth;
-    eauth.key.create(g_ceph_context, CEPH_CRYPTO_AES);
-    keyring->add(ename, eauth);
-    bufferlist bl;
-    keyring->encode_plaintext(bl);
-    int r = bl.write_file(g_conf->keyring.c_str(), 0600);
-    if (r)
-      derr << TEXT_RED << " ** ERROR: writing new keyring to " << g_conf->keyring
-	   << ": " << cpp_strerror(r) << TEXT_NORMAL << dendl;
-    else
-      derr << "created new key in keyring " << g_conf->keyring << dendl;
+
+    int ret = keyring->load(g_ceph_context, g_conf->keyring);
+    if (ret == 0 &&
+	keyring->get_auth(ename, eauth)) {
+      derr << "already have key in keyring " << g_conf->keyring << dendl;
+    } else {
+      eauth.key.create(g_ceph_context, CEPH_CRYPTO_AES);
+      keyring->add(ename, eauth);
+      bufferlist bl;
+      keyring->encode_plaintext(bl);
+      int r = bl.write_file(g_conf->keyring.c_str(), 0600);
+      if (r)
+	derr << TEXT_RED << " ** ERROR: writing new keyring to " << g_conf->keyring
+	     << ": " << cpp_strerror(r) << TEXT_NORMAL << dendl;
+      else
+	derr << "created new key in keyring " << g_conf->keyring << dendl;
+    }
   }
   if (mkfs || mkkey)
     exit(0);
@@ -251,6 +262,7 @@ int main(int argc, const char **argv)
 	   << ": " << cpp_strerror(-err) << TEXT_NORMAL << dendl;
       exit(1);
     }
+    exit(0);
   }
   
   string magic;
@@ -318,8 +330,6 @@ int main(int argc, const char **argv)
   messenger_hbclient->set_cluster_protocol(CEPH_OSD_PROTOCOL);
   messenger_hbserver->set_cluster_protocol(CEPH_OSD_PROTOCOL);
 
-  global_print_banner();
-
   cout << "starting osd." << whoami
        << " at " << client_messenger->get_myaddr()
        << " osd_data " << g_conf->osd_data
@@ -336,20 +346,18 @@ int main(int argc, const char **argv)
     CEPH_FEATURE_PGID64;
 
   client_messenger->set_default_policy(Messenger::Policy::stateless_server(supported, 0));
-  client_messenger->set_policy(entity_name_t::TYPE_CLIENT,
-			       Messenger::Policy::stateless_server(supported, 0));
-  client_messenger->set_policy_throttler(entity_name_t::TYPE_CLIENT, &client_throttler);
+  client_messenger->set_policy_throttler(entity_name_t::TYPE_CLIENT, &client_throttler);  // default, actually
   client_messenger->set_policy(entity_name_t::TYPE_MON,
-                               Messenger::Policy::client(supported,
-							 CEPH_FEATURE_UID |
-							 CEPH_FEATURE_PGID64 |
-							 CEPH_FEATURE_OSDENC));
+                               Messenger::Policy::lossy_client(supported,
+							       CEPH_FEATURE_UID |
+							       CEPH_FEATURE_PGID64 |
+							       CEPH_FEATURE_OSDENC));
   //try to poison pill any OSD connections on the wrong address
   client_messenger->set_policy(entity_name_t::TYPE_OSD,
 			       Messenger::Policy::stateless_server(0,0));
   
   cluster_messenger->set_default_policy(Messenger::Policy::stateless_server(0, 0));
-  cluster_messenger->set_policy(entity_name_t::TYPE_MON, Messenger::Policy::client(0,0));
+  cluster_messenger->set_policy(entity_name_t::TYPE_MON, Messenger::Policy::lossy_client(0,0));
   cluster_messenger->set_policy(entity_name_t::TYPE_OSD,
 				Messenger::Policy::lossless_peer(supported,
 								 CEPH_FEATURE_UID |
@@ -359,7 +367,7 @@ int main(int argc, const char **argv)
 				Messenger::Policy::stateless_server(0, 0));
 
   messenger_hbclient->set_policy(entity_name_t::TYPE_OSD,
-			     Messenger::Policy::client(0, 0));
+			     Messenger::Policy::lossy_client(0, 0));
   messenger_hbserver->set_policy(entity_name_t::TYPE_OSD,
 			     Messenger::Policy::stateless_server(0, 0));
 
@@ -381,16 +389,18 @@ int main(int argc, const char **argv)
   if (r < 0)
     exit(1);
 
+
   // Set up crypto, daemonize, etc.
-  // Leave stderr open in case we need to report errors.
-  global_init_daemonize(g_ceph_context, CINIT_FLAG_NO_CLOSE_STDERR);
+  global_init_daemonize(g_ceph_context, 0);
   common_init_finish(g_ceph_context);
 
-  int err = OSD::convertfs(g_conf->osd_data, g_conf->osd_journal);
-  if (err < 0) {
-    derr << TEXT_RED << " ** ERROR: error converting store " << g_conf->osd_data
-	 << ": " << cpp_strerror(-err) << TEXT_NORMAL << dendl;
-    exit(1);
+  if (g_conf->filestore_update_to >= (int)FileStore::on_disk_version) {
+    int err = OSD::convertfs(g_conf->osd_data, g_conf->osd_journal);
+    if (err < 0) {
+      derr << TEXT_RED << " ** ERROR: error converting store " << g_conf->osd_data
+	   << ": " << cpp_strerror(-err) << TEXT_NORMAL << dendl;
+      exit(1);
+    }
   }
 
   MonClient mc(g_ceph_context);
@@ -402,7 +412,8 @@ int main(int argc, const char **argv)
 		messenger_hbclient, messenger_hbserver,
 		&mc,
 		g_conf->osd_data, g_conf->osd_journal);
-  err = osd->pre_init();
+
+  int err = osd->pre_init();
   if (err < 0) {
     derr << TEXT_RED << " ** ERROR: osd pre_init failed: " << cpp_strerror(-err)
 	 << TEXT_NORMAL << dendl;
