@@ -43,6 +43,8 @@ using namespace __gnu_cxx;
 #include "common/Mutex.h"
 #include "common/Timer.h"
 
+#include "common/compiler_extensions.h"
+
 #include "osdc/ObjectCacher.h"
 
 class MDSMap;
@@ -215,6 +217,7 @@ public:
   list<Cond*> waiting_for_mdsmap;
 
   void got_mds_push(int mds);
+  void _closed_mds_session(int mds, MetaSession *s);
   void handle_client_session(MClientSession *m);
   void send_reconnect(int mds);
   void resend_unsafe_requests(int mds);
@@ -255,7 +258,6 @@ public:
   int file_stripe_count;
   int object_size;
   int file_replication;
-  int preferred_pg;
 public:
   entity_name_t get_myname() { return messenger->get_myname(); } 
   void sync_write_commit(Inode *in);
@@ -339,7 +341,7 @@ protected:
   void trim_dentry(Dentry *dn);
   void trim_caps(int mds, int max);
   
-  void dump_inode(Inode *in, set<Inode*>& did);
+  void dump_inode(Inode *in, set<Inode*>& did, bool disconnected);
   void dump_cache();  // debug
   
   // trace generation
@@ -371,7 +373,7 @@ protected:
 
   inodeno_t get_root_ino();
 
-  int init();
+  int init()  WARN_UNUSED_RESULT;
   void shutdown();
 
   // messaging
@@ -619,7 +621,6 @@ public:
   void set_default_file_stripe_count(int count);
   void set_default_object_size(int size);
   void set_default_file_replication(int replication);
-  void set_default_preferred_pg(int osd);
 
   int enumerate_layout(int fd, vector<ObjectExtent>& result,
 		       loff_t length, loff_t offset);
