@@ -67,7 +67,8 @@ void AuthMonitor::check_rotate()
 
 void AuthMonitor::tick() 
 {
-  if (!paxos->is_active()) return;
+  if (!paxos->is_active() ||
+      !mon->is_all_paxos_recovered()) return;
 
   update_from_paxos();
   dout(10) << *this << dendl;
@@ -425,8 +426,9 @@ done:
 void AuthMonitor::auth_usage(stringstream& ss)
 {
   ss << "error: usage:" << std::endl;
-  ss << "              auth <add | del> <name> <--in-file=filename>" << std::endl;
-  ss << "              auth <list>" << std::endl;
+  ss << "              auth (add | del | get-or-create | get-or-create-key | caps) <name> <--in-file=filename>" << std::endl;
+  ss << "              auth (export | get | get-key | print-key) <name>" << std::endl;
+  ss << "              auth list" << std::endl;
 }
 
 bool AuthMonitor::preprocess_command(MMonCommand *m)
