@@ -12,6 +12,8 @@
  * 
  */
 
+#include <sstream>
+#include <syslog.h>
 
 #include "LogMonitor.h"
 #include "Monitor.h"
@@ -24,10 +26,9 @@
 #include "common/Timer.h"
 
 #include "osd/osd_types.h"
-#include "osd/PG.h"  // yuck
-
+#include "common/errno.h"
 #include "common/config.h"
-#include <sstream>
+#include "include/assert.h"
 
 #define dout_subsys ceph_subsys_mon
 #undef dout_prefix
@@ -67,7 +68,8 @@ ostream& operator<<(ostream& out, LogMonitor& pm)
 
 void LogMonitor::tick() 
 {
-  if (!paxos->is_active()) return;
+  if (!paxos->is_active() ||
+      !mon->is_all_paxos_recovered()) return;
 
   update_from_paxos();
   dout(10) << *this << dendl;
@@ -98,11 +100,6 @@ void LogMonitor::update_from_paxos()
   assert(paxosv >= summary.version);
 
   bufferlist blog;
-  bufferlist blogdebug;
-  bufferlist bloginfo;
-  bufferlist blogwarn;
-  bufferlist blogerr;
-  bufferlist blogsec;
 
   if (summary.version != paxos->get_stashed_version()) {
     bufferlist latest;
@@ -128,21 +125,14 @@ void LogMonitor::update_from_paxos()
 
       stringstream ss;
       ss << le;
-      string s;
-      getline(ss, s);
-      s += "\n";
+      string s = ss.str();
 
-      blog.append(s);
-      if (le.type >= CLOG_DEBUG)
-	blogdebug.append(s);
-      if (le.type >= CLOG_INFO)
-	bloginfo.append(s);
-      if (le.type == CLOG_SEC)
-        blogsec.append(s);
-      if (le.type >= CLOG_WARN)
-	blogwarn.append(s);
-      if (le.type >= CLOG_ERROR)
-	blogerr.append(s);
+      if (g_conf->mon_cluster_log_to_syslog) {
+	syslog(clog_type_to_syslog_prio(le.type) | LOG_USER, "%s", s.c_str());
+      }
+      if (g_conf->mon_cluster_log_file.length()) {
+	blog.append(s + "\n");
+      }
 
       summary.add(le);
     }
@@ -154,24 +144,23 @@ void LogMonitor::update_from_paxos()
   ::encode(summary, bl);
   paxos->stash_latest(paxosv, bl);
 
-  if (blog.length())
-    mon->store->append_bl_ss(blog, "log", NULL);
-  if (blogdebug.length())
-    mon->store->append_bl_ss(blogdebug, "log.debug", NULL);
-  if (bloginfo.length())
-    mon->store->append_bl_ss(bloginfo, "log.info", NULL);
-  if (blogsec.length())
-    mon->store->append_bl_ss(bloginfo, "log.security", NULL);
-  if (blogwarn.length())
-    mon->store->append_bl_ss(blogwarn, "log.warn", NULL);
-  if (blogerr.length())
-    mon->store->append_bl_ss(blogerr, "log.err", NULL);
-
+  if (blog.length()) {
+    int fd = ::open(g_conf->mon_cluster_log_file.c_str(), O_WRONLY|O_APPEND|O_CREAT, 0600);
+    if (fd < 0) {
+      int err = -errno;
+      dout(1) << "unable to write to " << g_conf->mon_cluster_log_file << ": " << cpp_strerror(err) << dendl;
+    } else {
+      blog.write_fd(fd);
+      TEMP_FAILURE_RETRY(::close(fd));
+    }
+  }
 
   // trim
   unsigned max = g_conf->mon_max_log_epochs;
   if (mon->is_leader() && paxosv > max)
     paxos->trim_to(paxosv - max);
+
+  check_subs();
 }
 
 void LogMonitor::create_pending()
@@ -261,7 +250,8 @@ bool LogMonitor::prepare_log(MLog *m)
   dout(10) << "prepare_log " << *m << " from " << m->get_orig_source() << dendl;
 
   if (m->fsid != mon->monmap->fsid) {
-    dout(0) << "handle_log on fsid " << m->fsid << " != " << mon->monmap->fsid << dendl;
+    dout(0) << "handle_log on fsid " << m->fsid << " != " << mon->monmap->fsid 
+	    << dendl;
     m->put();
     return false;
   }
@@ -273,9 +263,9 @@ bool LogMonitor::prepare_log(MLog *m)
     if (!pending_summary.contains(p->key())) {
       pending_summary.add(*p);
       pending_log.insert(pair<utime_t,LogEntry>(p->stamp, *p));
+
     }
   }
-
   paxos->wait_for_commit(new C_Log(this, m));
   return true;
 }
@@ -284,9 +274,20 @@ void LogMonitor::_updated_log(MLog *m)
 {
   dout(7) << "_updated_log for " << m->get_orig_source_inst() << dendl;
   mon->send_reply(m, new MLogAck(m->fsid, m->entries.rbegin()->seq));
+
   m->put();
 }
 
+bool LogMonitor::should_propose(double& delay)
+{
+  // commit now if we have a lot of pending events
+  if (g_conf->mon_max_log_entries_per_event > 0 &&
+      pending_log.size() >= (unsigned)g_conf->mon_max_log_entries_per_event)
+    return true;
+
+  // otherwise fall back to generic policy
+  return PaxosService::should_propose(delay);
+}
 
 
 bool LogMonitor::preprocess_command(MMonCommand *m)
@@ -318,3 +319,159 @@ bool LogMonitor::prepare_command(MMonCommand *m)
   mon->reply_command(m, err, rs, paxos->get_version());
   return false;
 }
+
+
+int LogMonitor::sub_name_to_id(const string& n)
+{
+  if (n == "log-debug")
+    return CLOG_DEBUG;
+  if (n == "log-info")
+    return CLOG_INFO;
+  if (n == "log-sec")
+    return CLOG_SEC;
+  if (n == "log-warn")
+    return CLOG_WARN;
+  if (n == "log-error")
+    return CLOG_ERROR;
+  return -1;
+}
+
+void LogMonitor::check_subs()
+{
+  dout(10) << __func__ << dendl;
+  for (map<string, xlist<Subscription*>*>::iterator i = mon->session_map.subs.begin();
+       i != mon->session_map.subs.end();
+       i++) {
+    for (xlist<Subscription*>::iterator j = i->second->begin(); !j.end(); ++j) {
+      if (sub_name_to_id((*j)->type) >= 0)
+	check_sub(*j);
+    }
+  }
+}
+
+void LogMonitor::check_sub(Subscription *s)
+{
+  dout(10) << __func__ << " client wants " << s->type << " ver " << s->next << dendl;
+
+  int sub_level = sub_name_to_id(s->type);
+  assert(sub_level >= 0);
+
+  version_t summary_version = summary.version;
+  if (s->next > summary_version) {
+    dout(10) << __func__ << " client " << s->session->inst 
+	    << " requested version (" << s->next << ") is greater than ours (" 
+	    << summary_version << "), which means we already sent him" 
+	    << " everything we have." << dendl;
+    return;
+  } 
+ 
+  MLog *mlog = new MLog(mon->monmap->fsid);
+
+  if (s->next == 0) { 
+    /* First timer, heh? */
+    bool ret = _create_sub_summary(mlog, sub_level);
+    if (!ret) {
+      dout(1) << __func__ << " ret = " << ret << dendl;
+      mlog->put();
+      return;
+    }
+  } else {
+    /* let us send you an incremental log... */
+    _create_sub_incremental(mlog, sub_level, s->next);
+  }
+
+  dout(1) << __func__ << " sending message to " << s->session->inst 
+	  << " with " << mlog->entries.size() << " entries"
+	  << " (version " << mlog->version << ")" << dendl;
+  
+  mon->messenger->send_message(mlog, s->session->inst);
+  if (s->onetime)
+    mon->session_map.remove_sub(s);
+  else
+    s->next = summary_version+1;
+}
+
+/**
+ * Create a log message containing only the last message in the summary.
+ *
+ * @param mlog	Log message we'll send to the client.
+ * @param level Maximum log level the client is interested in.
+ * @return	'true' if we consider we successfully populated @mlog;
+ *		'false' otherwise.
+ */
+bool LogMonitor::_create_sub_summary(MLog *mlog, int level)
+{
+  dout(10) << __func__ << dendl;
+
+  assert(mlog != NULL);
+
+  if (!summary.tail.size())
+    return false;
+
+  list<LogEntry>::reverse_iterator it = summary.tail.rbegin();
+  for (; it != summary.tail.rend(); it++) {
+    LogEntry e = *it;
+    if (e.type < level)
+      continue;
+
+    mlog->entries.push_back(e);
+    mlog->version = summary.version;
+    break;
+  }
+
+  return true;
+}
+
+/**
+ * Create an incremental log message from version @sv to @summary.version
+ *
+ * @param mlog	Log message we'll send to the client with the messages received
+ *		since version @sv, inclusive.
+ * @param level	The max log level of the messages the client is interested in.
+ * @param sv	The version the client is looking for.
+ */
+void LogMonitor::_create_sub_incremental(MLog *mlog, int level, version_t sv)
+{
+  dout(10) << __func__ << " level " << level << " ver " << sv 
+	  << " cur summary ver " << summary.version << dendl; 
+
+  if (sv < paxos->get_first_committed()) {
+    dout(10) << __func__ << " skipped from " << sv
+	     << " to first_committed " << paxos->get_first_committed() << dendl;
+    LogEntry le;
+    le.stamp = ceph_clock_now(NULL);
+    le.type = CLOG_WARN;
+    ostringstream ss;
+    ss << "skipped log messages from " << sv << " to " << paxos->get_first_committed();
+    le.msg = ss.str();
+    mlog->entries.push_back(le);
+    sv = paxos->get_first_committed();
+  }
+
+  version_t summary_ver = summary.version;
+  while (sv <= summary_ver) {
+    bufferlist bl;
+    bool success = paxos->read(sv, bl);
+    assert(success);
+    bufferlist::iterator p = bl.begin();
+    __u8 v;
+    ::decode(v,p);
+    while (!p.end()) {
+      LogEntry le;
+      le.decode(p);
+
+      if (le.type < level) {
+	dout(20) << __func__ << " requested " << level 
+		 << " entry " << le.type << dendl;
+	continue;
+      }
+
+      mlog->entries.push_back(le);
+    }
+    mlog->version = sv++;
+  }
+
+  dout(10) << __func__ << " incremental message ready (" 
+	   << mlog->entries.size() << " entries)" << dendl;
+}
+
