@@ -45,6 +45,7 @@
 
 #include "include/compat.h"
 #include "include/assert.h"
+#include "include/stringify.h"
 
 #define dout_subsys ceph_subsys_mon
 #undef dout_prefix
@@ -672,74 +673,135 @@ bool OSDMonitor::can_mark_in(int i)
   return true;
 }
 
+void OSDMonitor::check_failures(utime_t now)
+{
+  for (map<int,failure_info_t>::iterator p = failure_info.begin();
+       p != failure_info.end();
+       ++p) {
+    check_failure(now, p->first, p->second);
+  }
+}
+
+bool OSDMonitor::check_failure(utime_t now, int target_osd, failure_info_t& fi)
+{
+  utime_t orig_grace(g_conf->osd_heartbeat_grace, 0);
+  utime_t max_failed_since = fi.get_failed_since();
+  utime_t failed_for = now - max_failed_since;
+
+  utime_t grace = orig_grace;
+  double my_grace = 0, peer_grace = 0;
+  if (g_conf->mon_osd_adjust_heartbeat_grace) {
+    double halflife = (double)g_conf->mon_osd_laggy_halflife;
+    double decay_k = ::log(.5) / halflife;
+
+    // scale grace period based on historical probability of 'lagginess'
+    // (false positive failures due to slowness).
+    const osd_xinfo_t& xi = osdmap.get_xinfo(target_osd);
+    double decay = exp((double)failed_for * decay_k);
+    dout(20) << " halflife " << halflife << " decay_k " << decay_k
+	     << " failed_for " << failed_for << " decay " << decay << dendl;
+    my_grace = decay * (double)xi.laggy_interval * xi.laggy_probability;
+    grace += my_grace;
+
+    // consider the peers reporting a failure a proxy for a potential
+    // 'subcluster' over the overall cluster that is similarly
+    // laggy.  this is clearly not true in all cases, but will sometimes
+    // help us localize the grace correction to a subset of the system
+    // (say, a rack with a bad switch) that is unhappy.
+    assert(fi.reporters.size());
+    for (map<int,failure_reporter_t>::iterator p = fi.reporters.begin();
+	 p != fi.reporters.end();
+	 p++) {
+      const osd_xinfo_t& xi = osdmap.get_xinfo(p->first);
+      utime_t elapsed = now - xi.down_stamp;
+      double decay = exp((double)elapsed * decay_k);
+      peer_grace += decay * (double)xi.laggy_interval * xi.laggy_probability;
+    }
+    peer_grace /= (double)fi.reporters.size();
+    grace += peer_grace;
+  }
+
+  dout(10) << " osd." << target_osd << " has "
+	   << fi.reporters.size() << " reporters and "
+	   << fi.num_reports << " reports, "
+	   << grace << " grace (" << orig_grace << " + " << my_grace << " + " << peer_grace << "), max_failed_since " << max_failed_since
+	   << dendl;
+
+  if (failed_for >= grace &&
+      ((int)fi.reporters.size() >= g_conf->osd_min_down_reporters) &&
+      (fi.num_reports >= g_conf->osd_min_down_reports)) {
+    dout(1) << " we have enough reports/reporters to mark osd." << target_osd << " down" << dendl;
+    pending_inc.new_state[target_osd] = CEPH_OSD_UP;
+
+    mon->clog.info() << osdmap.get_inst(target_osd) << " failed ("
+		     << fi.num_reports << " reports from " << (int)fi.reporters.size() << " peers after "
+		     << failed_for << " >= grace " << grace << ")\n";
+
+    list<MOSDFailure*> msgs;
+    fi.take_report_messages(msgs);
+    failure_info.erase(target_osd);
+
+    paxos->wait_for_commit(new C_Reported(this, msgs));
+    return true;
+  }
+  return false;
+}
+
 bool OSDMonitor::prepare_failure(MOSDFailure *m)
 {
   dout(1) << "prepare_failure " << m->get_target() << " from " << m->get_orig_source_inst()
           << " is reporting failure:" << m->if_osd_failed() << dendl;
-  mon->clog.info() << m->get_target() << " failed (by "
-		   << m->get_orig_source_inst() << ")\n";
+
+  mon->clog.debug() << m->get_target() << " reported failed by "
+		    << m->get_orig_source_inst() << "\n";
   
   int target_osd = m->get_target().name.num();
   int reporter = m->get_orig_source().num();
   assert(osdmap.is_up(target_osd));
   assert(osdmap.get_addr(target_osd) == m->get_target().addr);
+
+  // calculate failure time
+  utime_t now = ceph_clock_now(g_ceph_context);
+  utime_t failed_since = now - utime_t(m->failed_for ? m->failed_for : g_conf->osd_heartbeat_grace, 0);
   
   if (m->if_osd_failed()) {
-    int reports = 0;
-    int reporters = 0;
+    // add a report
+    failure_info_t& fi = failure_info[target_osd];
+    MOSDFailure *old = fi.add_report(reporter, failed_since, m);
+    if (old)
+      mon->no_reply(old);
 
-    if (failed_notes.count(target_osd)) {
-      multimap<int, pair<int, int> >::iterator i = failed_notes.lower_bound(target_osd);
-      while ((i != failed_notes.end()) && (i->first == target_osd)) {
-        if (i->second.first == reporter) {
-          ++i->second.second;
-          dout(10) << "adding new failure report from osd." << reporter
-                   << " on osd." << target_osd << dendl;
-          reporter = -1;
-        }
-        ++reporters;
-        reports += i->second.second;
-        ++i;
+    return check_failure(now, target_osd, fi);
+  } else {
+    // remove the report
+    if (failure_info.count(target_osd)) {
+      failure_info_t& fi = failure_info[target_osd];
+      fi.cancel_report(reporter);
+      if (fi.reporters.empty()) {
+	dout(10) << " removing last failure_info for osd." << target_osd << dendl;
+	failure_info.erase(target_osd);
+      } else {
+	dout(10) << " failure_info for osd." << target_osd << " now "
+		 << fi.reporters.size() << " reporters and "
+		 << fi.num_reports << " reports" << dendl;
       }
+    } else {
+      dout(10) << " no failure_info for osd." << target_osd << dendl;
     }
-    if (reporter != -1) { //didn't get counted yet
-      failed_notes.insert(pair<int, pair<int, int> >
-                          (target_osd, pair<int, int>(reporter, 1)));
-      ++reporters;
-      ++reports;
-      dout(10) << "osd." << reporter
-               << " is adding failure report on osd." << target_osd << dendl;
-    }
-
-    if ((reporters >= g_conf->osd_min_down_reporters) &&
-        (reports >= g_conf->osd_min_down_reports)) {
-      dout(1) << "have enough reports/reporters to mark osd." << target_osd
-              << " as down" << dendl;
-      pending_inc.new_state[target_osd] = CEPH_OSD_UP;
-      paxos->wait_for_commit(new C_Reported(this, m));
-      //clear out failure reports
-      failed_notes.erase(failed_notes.lower_bound(target_osd),
-                         failed_notes.upper_bound(target_osd));
-      return true;
-    }
-  } else { //remove the report
-    multimap<int, pair<int, int> >::iterator i = failed_notes.lower_bound(target_osd);
-    while ((i != failed_notes.end()) && (i->first == target_osd)
-                                && (i->second.first != reporter))
-      ++i;
-    if ((i == failed_notes.end()) || (i->second.first != reporter))
-      dout(0) << "got an OSD not-failed report from osd." << reporter
-              << " that hasn't reported failure! (or in previous epoch?)" << dendl;
-    else failed_notes.erase(i);
+    mon->no_reply(m);
   }
   
   return false;
 }
 
-void OSDMonitor::_reported_failure(MOSDFailure *m)
+void OSDMonitor::_reported_failure(list<MOSDFailure*>& ls)
 {
-  dout(7) << "_reported_failure on " << m->get_target() << ", telling " << m->get_orig_source_inst() << dendl;
-  send_latest(m, m->get_epoch());
+  dout(7) << "_reported_failure telling " << ls << dendl;
+  while (!ls.empty()) {
+    MOSDFailure *m = ls.front();
+    ls.pop_front();
+    send_latest(m, m->get_epoch());
+  }
 }
 
 
@@ -885,6 +947,25 @@ bool OSDMonitor::prepare_boot(MOSDBoot *m)
       pending_inc.new_last_clean_interval[from] = pair<epoch_t,epoch_t>(begin, end);
     }
 
+    osd_xinfo_t xi = osdmap.get_xinfo(from);
+    if (m->boot_epoch == 0) {
+      xi.laggy_probability *= (1.0 - g_conf->mon_osd_laggy_weight);
+      xi.laggy_interval *= (1.0 - g_conf->mon_osd_laggy_weight);
+      dout(10) << " not laggy, new xi " << xi << dendl;
+    } else {
+      if (xi.down_stamp.sec()) {
+	int interval = ceph_clock_now(g_ceph_context).sec() - xi.down_stamp.sec();
+	xi.laggy_interval =
+	  interval * g_conf->mon_osd_laggy_weight +
+	  xi.laggy_interval * (1.0 - g_conf->mon_osd_laggy_weight);
+      }
+      xi.laggy_probability =
+	g_conf->mon_osd_laggy_weight +
+	xi.laggy_probability * (1.0 - g_conf->mon_osd_laggy_weight);
+      dout(10) << " laggy, now xi " << xi << dendl;
+    }
+    pending_inc.new_xinfo[from] = xi;
+
     // wait
     paxos->wait_for_commit(new C_Booted(this, m));
   }
@@ -1288,7 +1369,8 @@ void OSDMonitor::check_sub(Subscription *sub)
 
 void OSDMonitor::tick()
 {
-  if (!paxos->is_active()) return;
+  if (!paxos->is_active() ||
+      !mon->is_all_paxos_recovered()) return;
 
   update_from_paxos();
   dout(10) << osdmap << dendl;
@@ -1296,10 +1378,12 @@ void OSDMonitor::tick()
   if (!mon->is_leader()) return;
 
   bool do_propose = false;
-
-  // mark down osds out?
   utime_t now = ceph_clock_now(g_ceph_context);
 
+  // mark osds down?
+  check_failures(now);
+
+  // mark down osds out?
 
   /* can_mark_out() checks if we can mark osds as being out. The -1 has no
    * influence at all. The decision is made based on the ratio of "in" osds,
@@ -1317,10 +1401,26 @@ void OSDMonitor::tick()
       if (osdmap.is_down(o) &&
 	  osdmap.is_in(o) &&
 	  can_mark_out(o)) {
+	utime_t orig_grace(g_conf->mon_osd_down_out_interval, 0);
+	utime_t grace = orig_grace;
+	double my_grace = 0.0;
+
+	if (g_conf->mon_osd_adjust_down_out_interval) {
+	  // scale grace period the same way we do the heartbeat grace.
+	  const osd_xinfo_t& xi = osdmap.get_xinfo(o);
+	  double halflife = (double)g_conf->mon_osd_laggy_halflife;
+	  double decay_k = ::log(.5) / halflife;
+	  double decay = exp((double)down * decay_k);
+	  dout(20) << "osd." << o << " laggy halflife " << halflife << " decay_k " << decay_k
+		   << " down for " << down << " decay " << decay << dendl;
+	  my_grace = decay * (double)xi.laggy_interval * xi.laggy_probability;
+	  grace += my_grace;
+	}
+
 	if (g_conf->mon_osd_down_out_interval > 0 &&
-	    down.sec() >= g_conf->mon_osd_down_out_interval) {
+	    down.sec() >= grace) {
 	  dout(10) << "tick marking osd." << o << " OUT after " << down
-		   << " sec (target " << g_conf->mon_osd_down_out_interval << ")" << dendl;
+		   << " sec (target " << grace << " = " << orig_grace << " + " << my_grace << ")" << dendl;
 	  pending_inc.new_weight[o] = CEPH_OSD_OUT;
 
 	  // set the AUTOOUT bit.
@@ -1536,9 +1636,14 @@ bool OSDMonitor::preprocess_command(MMonCommand *m)
 	  break;
 	else if (ceph_argparse_witharg(args, i, &val, "-f", "--format", (char*)NULL))
 	  format = val;
-	else if (!epoch)
-	  epoch = atoi(*i++);
-	else
+	else if (!epoch) {
+	  long l = parse_pos_long(*i++, &ss);
+	  if (l < 0) {
+	    r = -EINVAL;
+	    goto out;
+	  }
+	  epoch = l;
+	} else
 	  i++;
       }
 
@@ -1644,10 +1749,12 @@ bool OSDMonitor::preprocess_command(MMonCommand *m)
 	r = 0;
       }
     }
-    else if ((m->cmd[1] == "scrub" || m->cmd[1] == "repair")) {
+    else if ((m->cmd[1] == "scrub" ||
+	      m->cmd[1] == "deep-scrub" ||
+	      m->cmd[1] == "repair")) {
       if (m->cmd.size() <= 2) {
 	r = -EINVAL;
-	ss << "usage: osd [scrub|repair] <who>";
+	ss << "usage: osd [scrub|deep-scrub|repair] <who>";
 	goto out;
       }
       if (m->cmd[2] == "*") {
@@ -1657,7 +1764,8 @@ bool OSDMonitor::preprocess_command(MMonCommand *m)
 	  if (osdmap.is_up(i)) {
 	    ss << (c++ ? ",":"") << i;
 	    mon->try_send_message(new MOSDScrub(osdmap.get_fsid(),
-						m->cmd[1] == "repair"),
+						m->cmd[1] == "repair",
+						m->cmd[1] == "deep-scrub"),
 				  osdmap.get_inst(i));
 	  }	    
 	r = 0;
@@ -1666,7 +1774,8 @@ bool OSDMonitor::preprocess_command(MMonCommand *m)
 	long osd = strtol(m->cmd[2].c_str(), 0, 10);
 	if (osdmap.is_up(osd)) {
 	  mon->try_send_message(new MOSDScrub(osdmap.get_fsid(),
-					      m->cmd[1] == "repair"),
+					      m->cmd[1] == "repair",
+					      m->cmd[1] == "deep-scrub"),
 				osdmap.get_inst(osd));
 	  r = 0;
 	  ss << "osd." << osd << " instructed to " << m->cmd[1];
@@ -1785,6 +1894,43 @@ bool OSDMonitor::prepare_unset_flag(MMonCommand *m, int flag)
   return true;
 }
 
+int OSDMonitor::parse_osd_id(const char *s, stringstream *pss)
+{
+  // osd.NNN?
+  if (strncmp(s, "osd.", 4) == 0) {
+    s += 4;
+  }
+
+  // NNN?
+  ostringstream ss;
+  long id = parse_pos_long(s, &ss);
+  if (id < 0) {
+    *pss << ss.str();
+    return id;
+  }
+  if (id > 0xffff) {
+    *pss << "osd id " << id << " is too large";
+    return -ERANGE;
+  }
+  return id;
+}
+
+void OSDMonitor::parse_loc_map(const vector<string>& args, int start, map<string,string> *ploc)
+{
+  for (unsigned i = start; i < args.size(); ++i) {
+    const char *s = args[i].c_str();
+    const char *pos = strchr(s, '=');
+    if (!pos)
+      break;
+    string key(s, 0, pos-s);
+    string value(pos+1);
+    if (value.length())
+      (*ploc)[key] = value;
+    else
+      ploc->erase(key);
+  }
+}
+
 bool OSDMonitor::prepare_command(MMonCommand *m)
 {
   bool ret = false;
@@ -1836,31 +1982,34 @@ bool OSDMonitor::prepare_command(MMonCommand *m)
       paxos->wait_for_commit(new Monitor::C_Command(mon, m, 0, rs, paxos->get_version()));
       return true;
     }
-    else if (m->cmd.size() >= 6 && m->cmd[1] == "crush" && m->cmd[2] == "set") {
+    else if (m->cmd.size() >= 5 && m->cmd[1] == "crush" && m->cmd[2] == "set") {
       do {
-	// osd crush set <id> <name> <weight> [<loc1> [<loc2> ...]]
-	int id = atoi(m->cmd[3].c_str());
+	// osd crush set <osd-id> [<osd.* name>] <weight> <loc1> [<loc2> ...]
+	int id = parse_osd_id(m->cmd[3].c_str(), &ss);
+	if (id < 0) {
+	  err = -EINVAL;
+	  goto out;
+	}
 	if (!osdmap.exists(id)) {
 	  err = -ENOENT;
-	  ss << "osd." << id << " does not exist.  create it before updating the crush map";
+	  ss << "osd." << m->cmd[3] << " does not exist.  create it before updating the crush map";
 	  goto out;
 	}
 
-	string name = m->cmd[4];
-	float weight = atof(m->cmd[5].c_str());
-	map<string,string> loc;
-	for (unsigned i = 6; i < m->cmd.size(); ++i) {
-	  const char *s = m->cmd[i].c_str();
-	  const char *pos = strchr(s, '=');
-	  if (!pos)
-	    break;
-	  string key(s, 0, pos-s);
-	  string value(pos+1);
-	  if (value.length())
-	    loc[key] = value;
-	  else
-	    loc.erase(key);
+	int argpos = 4;
+	string name;
+	if (m->cmd[argpos].find("osd.") == 0) {
+	  // old annoying usage, explicitly specifying osd.NNN name
+	  name = m->cmd[argpos];
+	  argpos++;
+	} else {
+	  // new usage; infer name
+	  name = "osd." + stringify(id);
 	}
+	float weight = atof(m->cmd[argpos].c_str());
+	argpos++;
+	map<string,string> loc;
+	parse_loc_map(m->cmd, argpos, &loc);
 
 	dout(0) << "adding/updating crush item id " << id << " name '" << name << "' weight " << weight
 		<< " at location " << loc << dendl;
@@ -1891,23 +2040,60 @@ bool OSDMonitor::prepare_command(MMonCommand *m)
 	}
       } while (false);
     }
+    else if (m->cmd.size() >= 5 && m->cmd[1] == "crush" && m->cmd[2] == "create-or-move") {
+      do {
+	// osd crush create-or-move <id> <initial_weight> <loc1> [<loc2> ...]
+	int id = parse_osd_id(m->cmd[3].c_str(), &ss);
+	if (id < 0) {
+	  err = -EINVAL;
+	  goto out;
+	}
+	if (!osdmap.exists(id)) {
+	  err = -ENOENT;
+	  ss << "osd." << m->cmd[3] << " does not exist.  create it before updating the crush map";
+	  goto out;
+	}
+
+	string name = "osd." + stringify(id);
+	float weight = atof(m->cmd[4].c_str());
+	map<string,string> loc;
+	parse_loc_map(m->cmd, 5, &loc);
+
+	dout(0) << "create-or-move crush item id " << id << " name '" << name << "' initial_weight " << weight
+		<< " at location " << loc << dendl;
+	bufferlist bl;
+	if (pending_inc.crush.length())
+	  bl = pending_inc.crush;
+	else
+	  osdmap.crush->encode(bl);
+
+	CrushWrapper newcrush;
+	bufferlist::iterator p = bl.begin();
+	newcrush.decode(p);
+
+	err = newcrush.create_or_move_item(g_ceph_context, id, weight, name, loc);
+	if (err == 0) {
+	  ss << "create-or-move updated item id " << id << " name '" << name << "' weight " << weight
+	     << " at location " << loc << " to crush map";
+	  break;
+	}
+	if (err > 0) {
+	  pending_inc.crush.clear();
+	  newcrush.encode(pending_inc.crush);
+	  ss << "create-or-move updating item id " << id << " name '" << name << "' weight " << weight
+	     << " at location " << loc << " to crush map";
+	  getline(ss, rs);
+	  paxos->wait_for_commit(new Monitor::C_Command(mon, m, 0, rs, paxos->get_version()));
+	  return true;
+	}
+      } while (false);
+    }
     else if (m->cmd.size() >= 5 && m->cmd[1] == "crush" && m->cmd[2] == "move") {
       do {
-	// osd crush move <name> [<loc1> [<loc2> ...]]
+	// osd crush move <name> <loc1> [<loc2> ...]
 	string name = m->cmd[3];
 	map<string,string> loc;
-	for (unsigned i = 4; i < m->cmd.size(); ++i) {
-	  const char *s = m->cmd[i].c_str();
-	  const char *pos = strchr(s, '=');
-	  if (!pos)
-	    break;
-	  string key(s, 0, pos-s);
-	  string value(pos+1);
-	  if (value.length())
-	    loc[key] = value;
-	  else
-	    loc.erase(key);
-	}
+	parse_loc_map(m->cmd, 4, &loc);
 
 	dout(0) << "moving crush item name '" << name << "' to location " << loc << dendl;
 	bufferlist bl;
@@ -2004,7 +2190,11 @@ bool OSDMonitor::prepare_command(MMonCommand *m)
       } while (false);
     }
     else if (m->cmd[1] == "setmaxosd" && m->cmd.size() > 2) {
-      int newmax = atoi(m->cmd[2].c_str());
+      int newmax = parse_pos_long(m->cmd[2].c_str(), &ss);
+      if (newmax < 0) {
+	err = -EINVAL;
+	goto out;
+      }
       if (newmax < osdmap.crush->get_max_devices()) {
 	err = -ERANGE;
 	ss << "cannot set max_osd to " << newmax << " which is < crush max_devices "
@@ -2060,6 +2250,18 @@ bool OSDMonitor::prepare_command(MMonCommand *m)
     else if (m->cmd.size() == 3 && m->cmd[1] == "unset" && m->cmd[2] == "noin") {
       return prepare_unset_flag(m, CEPH_OSDMAP_NOIN);
     }
+    else if (m->cmd.size() == 3 && m->cmd[1] == "set" && m->cmd[2] == "nobackfill") {
+      return prepare_set_flag(m, CEPH_OSDMAP_NOBACKFILL);
+    }
+    else if (m->cmd.size() == 3 && m->cmd[1] == "unset" && m->cmd[2] == "nobackfill") {
+      return prepare_unset_flag(m, CEPH_OSDMAP_NOBACKFILL);
+    }
+    else if (m->cmd.size() == 3 && m->cmd[1] == "set" && m->cmd[2] == "norecover") {
+      return prepare_set_flag(m, CEPH_OSDMAP_NORECOVER);
+    }
+    else if (m->cmd.size() == 3 && m->cmd[1] == "unset" && m->cmd[2] == "norecover") {
+      return prepare_unset_flag(m, CEPH_OSDMAP_NORECOVER);
+    }
     else if (m->cmd[1] == "cluster_snap" && m->cmd.size() == 3) {
       // ** DISABLE THIS FOR NOW **
       ss << "cluster snapshot currently disabled (broken implementation)";
@@ -2536,7 +2738,11 @@ bool OSDMonitor::prepare_command(MMonCommand *m)
 	     (m->cmd[1] == "reweight-by-utilization")) {
       int oload = 120;
       if (m->cmd.size() > 2) {
-	oload = atoi(m->cmd[2].c_str());
+	oload = parse_pos_long(m->cmd[2].c_str(), &ss);
+	if (oload < 0) {
+	  err = -EINVAL;
+	  goto out;
+	}
       }
       string out_str;
       err = reweight_by_utilization(oload, out_str);
@@ -2553,7 +2759,12 @@ bool OSDMonitor::prepare_command(MMonCommand *m)
       }
     }
     else if (m->cmd.size() == 3 && m->cmd[1] == "thrash") {
-      thrash_map = atoi(m->cmd[2].c_str());
+      long l = parse_pos_long(m->cmd[2].c_str(), &ss);
+      if (l < 0) {
+	err = -EINVAL;
+	goto out;
+      }
+      thrash_map = l;
       ss << "will thrash map for " << thrash_map << " epochs";
       ret = thrash();
       err = 0;
