@@ -7,13 +7,28 @@
 #include "common/debug.h"
 
 #include "common/Formatter.h"
+#include "include/ceph_features.h"
 
 // --
 
-void PGMap::Incremental::encode(bufferlist &bl) const
+void PGMap::Incremental::encode(bufferlist &bl, uint64_t features) const
 {
-  __u8 v = 4;
-  ::encode(v, bl);
+  if ((features & CEPH_FEATURE_MONENC) == 0) {
+    __u8 v = 4;
+    ::encode(v, bl);
+    ::encode(version, bl);
+    ::encode(pg_stat_updates, bl);
+    ::encode(osd_stat_updates, bl);
+    ::encode(osd_stat_rm, bl);
+    ::encode(osdmap_epoch, bl);
+    ::encode(pg_scan, bl);
+    ::encode(full_ratio, bl);
+    ::encode(nearfull_ratio, bl);
+    ::encode(pg_remove, bl);
+    return;
+  }
+
+  ENCODE_START(5, 5, bl);
   ::encode(version, bl);
   ::encode(pg_stat_updates, bl);
   ::encode(osd_stat_updates, bl);
@@ -23,14 +38,14 @@ void PGMap::Incremental::encode(bufferlist &bl) const
   ::encode(full_ratio, bl);
   ::encode(nearfull_ratio, bl);
   ::encode(pg_remove, bl);
+  ENCODE_FINISH(bl);
 }
 
 void PGMap::Incremental::decode(bufferlist::iterator &bl)
 {
-  __u8 v;
-  ::decode(v, bl);
+  DECODE_START_LEGACY_COMPAT_LEN(5, 5, 5, bl);
   ::decode(version, bl);
-  if (v < 3) {
+  if (struct_v < 3) {
     pg_stat_updates.clear();
     __u32 n;
     ::decode(n, bl);
@@ -47,11 +62,11 @@ void PGMap::Incremental::decode(bufferlist::iterator &bl)
   ::decode(osd_stat_rm, bl);
   ::decode(osdmap_epoch, bl);
   ::decode(pg_scan, bl);
-  if (v >= 2) {
+  if (struct_v >= 2) {
     ::decode(full_ratio, bl);
     ::decode(nearfull_ratio, bl);
   }
-  if (v < 3) {
+  if (struct_v < 3) {
     pg_remove.clear();
     __u32 n;
     ::decode(n, bl);
@@ -63,12 +78,13 @@ void PGMap::Incremental::decode(bufferlist::iterator &bl)
   } else {
     ::decode(pg_remove, bl);
   }
-  if (v < 4 && full_ratio == 0) {
+  if (struct_v < 4 && full_ratio == 0) {
     full_ratio = -1;
   }
-  if (v < 4 && nearfull_ratio == 0) {
+  if (struct_v < 4 && nearfull_ratio == 0) {
     nearfull_ratio = -1;
   }
+  DECODE_FINISH(bl);
 }
 
 void PGMap::Incremental::dump(Formatter *f) const
@@ -173,29 +189,15 @@ void PGMap::apply_incremental(const Incremental& inc)
     if (t == osd_stat.end()) {
       hash_map<int,osd_stat_t>::value_type v(osd, new_stats);
       osd_stat.insert(v);
-    }
-    else {
+    } else {
       stat_osd_sub(t->second);
       t->second = new_stats;
     }
-    
+
     stat_osd_add(new_stats);
     
-    //update the full/nearful_osd sets
-    int from = p->first;
-    float ratio = ((float)p->second.kb_used) / (float) p->second.kb;
-    if ( ratio > full_ratio ) {
-      full_osds.insert(from);
-      //sets don't double-insert so this might be a (very expensive) null-op
-    }
-    else if ( ratio > nearfull_ratio ) {
-      nearfull_osds.insert(from);
-      full_osds.erase(from);
-    }
-    else {//it's not full or near-full
-      full_osds.erase(from);
-      nearfull_osds.erase(from);
-    }
+    // adjust [near]full status
+    register_nearfull_status(osd, new_stats);
   }
   for (set<pg_t>::const_iterator p = inc.pg_remove.begin();
        p != inc.pg_remove.end();
@@ -216,6 +218,10 @@ void PGMap::apply_incremental(const Incremental& inc)
       stat_osd_sub(t->second);
       osd_stat.erase(t);
     }
+
+    // remove these old osds from full/nearfull set(s), too
+    nearfull_osds.erase(*p);
+    full_osds.erase(*p);
   }
   
   if (inc.osdmap_epoch)
@@ -231,11 +237,26 @@ void PGMap::redo_full_sets()
   for (hash_map<int, osd_stat_t>::iterator i = osd_stat.begin();
        i != osd_stat.end();
        ++i) {
-    float ratio = ((float)i->second.kb_used) / ((float)i->second.kb);
-    if (full_ratio > 0 && ratio > full_ratio)
-      full_osds.insert(i->first);
-    else if (nearfull_ratio > 0 && ratio > nearfull_ratio)
-      nearfull_osds.insert(i->first);
+    register_nearfull_status(i->first, i->second);
+  }
+}
+
+void PGMap::register_nearfull_status(int osd, const osd_stat_t& s)
+{
+  float ratio = ((float)s.kb_used) / ((float)s.kb);
+
+  if (full_ratio > 0 && ratio > full_ratio) {
+    // full
+    full_osds.insert(osd);
+    nearfull_osds.erase(osd);
+  } else if (nearfull_ratio > 0 && ratio > nearfull_ratio) {
+    // nearfull
+    full_osds.erase(osd);
+    nearfull_osds.insert(osd);
+  } else {
+    // ok
+    full_osds.erase(osd);
+    nearfull_osds.erase(osd);
   }
 }
 
@@ -267,8 +288,11 @@ void PGMap::stat_pg_add(const pg_t &pgid, const pg_stat_t &s)
   num_pg_by_state[s.state]++;
   pg_pool_sum[pgid.pool()].add(s);
   pg_sum.add(s);
-  if (s.state & PG_STATE_CREATING)
+  if (s.state & PG_STATE_CREATING) {
     creating_pgs.insert(pgid);
+    if (s.acting.size())
+      creating_pgs_by_osd[s.acting[0]].insert(pgid);
+  }
 }
 
 void PGMap::stat_pg_sub(const pg_t &pgid, const pg_stat_t &s)
@@ -283,8 +307,11 @@ void PGMap::stat_pg_sub(const pg_t &pgid, const pg_stat_t &s)
     pg_pool_sum.erase(pgid.pool());
 
   pg_sum.sub(s);
-  if (s.state & PG_STATE_CREATING)
+  if (s.state & PG_STATE_CREATING) {
     creating_pgs.erase(pgid);
+    if (s.acting.size())
+      creating_pgs_by_osd[s.acting[0]].erase(pgid);
+  }
 }
 
 void PGMap::stat_osd_add(const osd_stat_t &s)
@@ -312,10 +339,22 @@ epoch_t PGMap::calc_min_last_epoch_clean() const
   return min;
 }
 
-void PGMap::encode(bufferlist &bl) const
+void PGMap::encode(bufferlist &bl, uint64_t features) const
 {
-  __u8 v = 3;
-  ::encode(v, bl);
+  if ((features & CEPH_FEATURE_MONENC) == 0) {
+    __u8 v = 3;
+    ::encode(v, bl);
+    ::encode(version, bl);
+    ::encode(pg_stat, bl);
+    ::encode(osd_stat, bl);
+    ::encode(last_osdmap_epoch, bl);
+    ::encode(last_pg_scan, bl);
+    ::encode(full_ratio, bl);
+    ::encode(nearfull_ratio, bl);
+    return;
+  }
+
+  ENCODE_START(4, 4, bl);
   ::encode(version, bl);
   ::encode(pg_stat, bl);
   ::encode(osd_stat, bl);
@@ -323,14 +362,14 @@ void PGMap::encode(bufferlist &bl) const
   ::encode(last_pg_scan, bl);
   ::encode(full_ratio, bl);
   ::encode(nearfull_ratio, bl);
+  ENCODE_FINISH(bl);
 }
 
 void PGMap::decode(bufferlist::iterator &bl)
 {
-  __u8 v;
-  ::decode(v, bl);
+  DECODE_START_LEGACY_COMPAT_LEN(4, 4, 4, bl);
   ::decode(version, bl);
-  if (v < 3) {
+  if (struct_v < 3) {
     pg_stat.clear();
     __u32 n;
     ::decode(n, bl);
@@ -346,10 +385,11 @@ void PGMap::decode(bufferlist::iterator &bl)
   ::decode(osd_stat, bl);
   ::decode(last_osdmap_epoch, bl);
   ::decode(last_pg_scan, bl);
-  if (v >= 2) {
+  if (struct_v >= 2) {
     ::decode(full_ratio, bl);
     ::decode(nearfull_ratio, bl);
   }
+  DECODE_FINISH(bl);
 
   calc_stats();
 }
@@ -424,7 +464,7 @@ void PGMap::dump_osd_stats(Formatter *f) const
 void PGMap::dump_pg_stats_plain(ostream& ss,
 				const hash_map<pg_t, pg_stat_t>& pg_stats) const
 {
-  ss << "pg_stat\tobjects\tmip\tdegr\tunf\tbytes\tlog\tdisklog\tstate\tstate_stamp\tv\treported\tup\tacting\tlast_scrub\tscrub_stamp" << std::endl;
+  ss << "pg_stat\tobjects\tmip\tdegr\tunf\tbytes\tlog\tdisklog\tstate\tstate_stamp\tv\treported\tup\tacting\tlast_scrub\tscrub_stamp\tlast_deep_scrub\tdeep_scrub_stamp" << std::endl;
   for (hash_map<pg_t, pg_stat_t>::const_iterator i = pg_stats.begin();
        i != pg_stats.end(); ++i) {
     const pg_stat_t &st(i->second);
@@ -444,6 +484,7 @@ void PGMap::dump_pg_stats_plain(ostream& ss,
        << "\t" << st.up
        << "\t" << st.acting
        << "\t" << st.last_scrub << "\t" << st.last_scrub_stamp
+       << "\t" << st.last_deep_scrub << "\t" << st.last_deep_scrub_stamp
        << std::endl;
   }
 }
