@@ -33,20 +33,40 @@ using namespace __gnu_cxx;
 
 #include "Messenger.h"
 #include "Message.h"
-#include "tcp.h"
 #include "include/assert.h"
+#include "DispatchQueue.h"
 
-
+#include "Pipe.h"
+#include "Accepter.h"
 
 /*
  * This class handles transmission and reception of messages. Generally
- * speaking, there are 2 major components:
- * 1) Pipe. Each network connection is handled through a pipe, which handles
- *    the input and output of each message.
- * 2) SimpleMessenger. It's the exterior class passed to the external 
- *    message handler and handles queuing and ordering of pipes. Each
- *    pipe maintains its own message ordering, but the SimpleMessenger
- *    decides what order pipes get to deliver messages in.
+ * speaking, there are several major components:
+ *
+ * - Connection
+ *    Each logical session is associated with a Connection.
+ * - Pipe
+ *    Each network connection is handled through a pipe, which handles
+ *    the input and output of each message.  There is normally a 1:1
+ *    relationship between Pipe and Connection, but logical sessions may
+ *    get handed off between Pipes when sockets reconnect or during
+ *    connection races.
+ * - IncomingQueue
+ *    Incoming messages are associated with an IncomingQueue, and there
+ *    is one such queue associated with each Pipe.
+ * - DispatchQueue
+ *    IncomingQueues get queued in the DIspatchQueue, which is responsible
+ *    for doing a round-robin sweep and processing them via a worker thread.
+ * - SimpleMessenger
+ *    It's the exterior class passed to the external message handler and
+ *    most of the API details.
+ *
+ * Lock ordering:
+ *
+ *   SimpleMessenger::lock
+ *       Pipe::pipe_lock
+ *           DispatchQueue::lock
+ *               IncomingQueue::lock
  */
 
 class SimpleMessenger : public Messenger {
@@ -61,39 +81,14 @@ public:
    * be a value that will be repeated if the daemon restarts.
    */
   SimpleMessenger(CephContext *cct, entity_name_t name,
-		  string mname, uint64_t _nonce) :
-    Messenger(cct, name),
-    accepter(this),
-    dispatch_queue(cct, this),
-    reaper_thread(this),
-    dispatch_thread(this),
-    my_type(name.type()),
-    nonce(_nonce),
-    lock("SimpleMessenger::lock"), need_addr(true), did_bind(false),
-    global_seq(0),
-    destination_stopped(false),
-    cluster_protocol(0),
-    dispatch_throttler(cct, string("msgr_dispatch_throttler-") + mname, cct->_conf->ms_dispatch_throttle_bytes),
-    reaper_started(false), reaper_stop(false),
-    timeout(0),
-    msgr(this)
-  {
-    pthread_spin_init(&global_seq_lock, PTHREAD_PROCESS_PRIVATE);
-    // for local dmsg delivery
-    dispatch_queue.local_pipe = new Pipe(this, Pipe::STATE_OPEN, NULL);
-    init_local_pipe();
-  }
+		  string mname, uint64_t _nonce);
+
   /**
    * Destroy the SimpleMessenger. Pretty simple since all the work is done
    * elsewhere.
    */
-  virtual ~SimpleMessenger() {
-    assert(destination_stopped); // we've been marked as stopped
-    assert(!did_bind); // either we didn't bind or we shut down the Accepter
-    assert(rank_pipe.empty()); // we don't have any running Pipes.
-    assert(reaper_stop && !reaper_started); // the reaper thread is stopped
-    delete dispatch_queue.local_pipe;
-  }
+  virtual ~SimpleMessenger();
+
   /** @defgroup Accessors
    * @{
    */
@@ -141,7 +136,7 @@ public:
    * @param p The Policy to apply.
    */
   void set_default_policy(Policy p) {
-    assert(!started && !did_bind);
+    Mutex::Locker l(policy_lock);
     default_policy = p;
   }
   /**
@@ -153,7 +148,7 @@ public:
    * @param p The policy to apply.
    */
   void set_policy(int type, Policy p) {
-    assert(!started && !did_bind);
+    Mutex::Locker l(policy_lock);
     policy_map[type] = p;
   }
   /**
@@ -168,9 +163,11 @@ public:
    * you destroy SimpleMessenger.
    */
   void set_policy_throttler(int type, Throttle *t) {
-    assert (!started && !did_bind);
-    assert(policy_map.count(type));
-    policy_map[type].throttler = t;
+    Mutex::Locker l(policy_lock);
+    if (policy_map.count(type))
+      policy_map[type].throttler = t;
+    else
+      default_policy.throttler = t;
   }
   /**
    * Bind the SimpleMessenger to a specific address. If bind_addr
@@ -387,401 +384,21 @@ private:
    * @defgroup Inner classes
    * @{
    */
-  /**
-   * If the SimpleMessenger binds to a specific address, the Accepter runs
-   * and listens for incoming connections.
-   */
-  class Accepter : public Thread {
-  public:
-    SimpleMessenger *msgr;
-    bool done;
-    int listen_sd;
-    
-    Accepter(SimpleMessenger *r) : msgr(r), done(false), listen_sd(-1) {}
-    
-    void *entry();
-    void stop();
-    int bind(entity_addr_t &bind_addr, int avoid_port1=0, int avoid_port2=0);
-    int rebind(int avoid_port);
-    int start();
-  } accepter;
 
+public:
+  Accepter accepter;
+  DispatchQueue dispatch_queue;
 
-  class DispatchQueue;
-  class Pipe;
-  struct IncomingQueue {
-    CephContext *cct;
-    Pipe *pipe;  // this will change
-    Mutex lock;
-    map<int, list<Message*> > in_q; // and inbound ones
-    int in_qlen;
-    map<int, xlist<IncomingQueue *>::item* > queue_items; // protected by pipe_lock AND q.lock
-    bool halt;
-
-    void queue(Message *m, int priority, DispatchQueue *dq);
-    void discard_queue(SimpleMessenger *msgr, DispatchQueue *dq);
-    void restart_queue();
-
-    IncomingQueue(CephContext *cct, Pipe *parent)
-      : cct(cct),
-	pipe(parent),
-	lock("SimpleMessenger::IncomingQueue::lock"),
-	in_qlen(0),
-	halt(false)
-    {
-    }
-    ~IncomingQueue() {
-      for (map<int, xlist<IncomingQueue *>::item* >::iterator i = queue_items.begin();
-	   i != queue_items.end();
-	   ++i) {
-        assert(!i->second->is_on_list());
-        delete i->second;
-      }
-    }
-  };
-
+  friend class Accepter;
 
   /**
-   * The Pipe is the most complex SimpleMessenger component. It gets
-   * two threads, one each for reading and writing on a socket it's handed
-   * at creation time, and is responsible for everything that happens on
-   * that socket. Besides message transmission, it's responsible for
-   * propagating socket errors to the SimpleMessenger and then sticking
-   * around in a state where it can provide enough data for the SimpleMessenger
-   * to provide reliable Message delivery when it manages to reconnect.
+   * Register a new pipe for accept
+   *
+   * @param sd socket
    */
-  class Pipe : public RefCountedObject {
-    /**
-     * The Reader thread handles all reads off the socket -- not just
-     * Messages, but also acks and other protocol bits (excepting startup,
-     * when the Writer does a couple of reads).
-     * All the work is implemented in Pipe itself, of course.
-     */
-    class Reader : public Thread {
-      Pipe *pipe;
-    public:
-      Reader(Pipe *p) : pipe(p) {}
-      void *entry() { pipe->reader(); return 0; }
-    } reader_thread;
-    friend class Reader;
+  Pipe *add_accept_pipe(int sd);
 
-    /**
-     * The Writer thread handles all writes to the socket (after startup).
-     * All the work is implemented in Pipe itself, of course.
-     */
-    class Writer : public Thread {
-      Pipe *pipe;
-    public:
-      Writer(Pipe *p) : pipe(p) {}
-      void *entry() { pipe->writer(); return 0; }
-    } writer_thread;
-    friend class Writer;
-
-  public:
-    Pipe(SimpleMessenger *r, int st, Connection *con) :
-      reader_thread(this), writer_thread(this),
-      msgr(r),
-      sd(-1),
-      peer_type(-1),
-      pipe_lock("SimpleMessenger::Pipe::pipe_lock"),
-      state(st),
-      connection_state(new Connection),
-      reader_running(false), reader_joining(false), writer_running(false),
-      in_q(new IncomingQueue(r->cct, this)),
-      keepalive(false),
-      close_on_empty(false),
-      connect_seq(0), peer_global_seq(0),
-      out_seq(0), in_seq(0), in_seq_acked(0) {
-      if (con) {
-        connection_state = con->get();
-        connection_state->reset_pipe(this);
-      } else {
-        connection_state = new Connection();
-        connection_state->pipe = get();
-      }
-      msgr->timeout = msgr->cct->_conf->ms_tcp_read_timeout * 1000; //convert to ms
-      if (msgr->timeout == 0)
-        msgr->timeout = -1;
-    }
-    ~Pipe() {
-      delete in_q;
-      assert(out_q.empty());
-      assert(sent.empty());
-      if (connection_state)
-        connection_state->put();
-    }
-
-    SimpleMessenger *msgr;
-    ostream& _pipe_prefix(std::ostream *_dout);
-
-    enum {
-      STATE_ACCEPTING,
-      STATE_CONNECTING,
-      STATE_OPEN,
-      STATE_STANDBY,
-      STATE_CLOSED,
-      STATE_CLOSING,
-      STATE_WAIT       // just wait for racing connection
-    };
-
-    int sd;
-    int peer_type;
-    entity_addr_t peer_addr;
-    Policy policy;
-    
-    Mutex pipe_lock;
-    int state;
-
-  protected:
-    friend class SimpleMessenger;
-    Connection *connection_state;
-
-    utime_t backoff;         // backoff time
-
-    bool reader_running, reader_joining;
-    bool writer_running;
-
-    map<int, list<Message*> > out_q;  // priority queue for outbound msgs
-    IncomingQueue *in_q;
-    list<Message*> sent;
-    Cond cond;
-    bool keepalive;
-    bool halt_delivery; //if a pipe's queue is destroyed, stop adding to it
-    bool close_on_empty;
-    
-    __u32 connect_seq, peer_global_seq;
-    uint64_t out_seq;
-    uint64_t in_seq, in_seq_acked;
-    
-    int accept();   // server handshake
-    int connect();  // client handshake
-    void reader();
-    void writer();
-    void unlock_maybe_reap();
-
-    int read_message(Message **pm);
-    int write_message(Message *m);
-    /**
-     * Write the given data (of length len) to the Pipe's socket. This function
-     * will loop until all passed data has been written out.
-     * If more is set, the function will optimize socket writes
-     * for additional data (by passing the MSG_MORE flag, aka TCP_CORK).
-     *
-     * @param msg The msghdr to write out
-     * @param len The length of the data in msg
-     * @param more Should be set true if this is one part of a larger message
-     * @return 0, or -1 on failure (unrecoverable -- close the socket).
-     */
-    int do_sendmsg(struct msghdr *msg, int len, bool more=false);
-    int write_ack(uint64_t s);
-    int write_keepalive();
-
-    void fault(bool onconnect=false, bool reader=false);
-    void fail();
-
-    void was_session_reset();
-
-    /* Clean up sent list */
-    void handle_ack(uint64_t seq) {
-      lsubdout(msgr->cct, ms, 15) << "reader got ack seq " << seq << dendl;
-      // trim sent list
-      while (!sent.empty() &&
-          sent.front()->get_seq() <= seq) {
-        Message *m = sent.front();
-        sent.pop_front();
-        lsubdout(msgr->cct, ms, 10) << "reader got ack seq "
-            << seq << " >= " << m->get_seq() << " on " << m << " " << *m << dendl;
-        m->put();
-      }
-
-      if (sent.empty() && close_on_empty) {
-        lsubdout(msgr->cct, ms, 10) << "reader got last ack, queue empty, closing" << dendl;
-        stop();
-      }
-    }
-
-    public:
-    Pipe(const Pipe& other);
-    const Pipe& operator=(const Pipe& other);
-
-    void start_reader() {
-      assert(pipe_lock.is_locked());
-      assert(!reader_running);
-      reader_running = true;
-      reader_thread.create(msgr->cct->_conf->ms_rwthread_stack_bytes);
-    }
-    void start_writer() {
-      assert(pipe_lock.is_locked());
-      assert(!writer_running);
-      writer_running = true;
-      writer_thread.create(msgr->cct->_conf->ms_rwthread_stack_bytes);
-    }
-    void join_reader() {
-      if (!reader_running)
-        return;
-      assert(!reader_joining);
-      reader_joining = true;
-      cond.Signal();
-      pipe_lock.Unlock();
-      reader_thread.join();
-      pipe_lock.Lock();
-      assert(reader_joining);
-      reader_joining = false;
-    }
-
-    // public constructors
-    static const Pipe& Server(int s);
-    static const Pipe& Client(const entity_addr_t& pi);
-
-    //we have two queue_received's to allow local signal delivery
-    // via Message * (that doesn't actually point to a Message)
-    void queue_received(Message *m, int priority);
-    
-    void queue_received(Message *m) {
-      // this is just to make sure that a changeset is working
-      // properly; if you start using the refcounting more and have
-      // multiple people hanging on to a message, ditch the assert!
-      assert(m->nref.read() == 1);
-
-      queue_received(m, m->get_priority());
-    }
-
-    __u32 get_out_seq() { return out_seq; }
-
-    bool is_queued() { return !out_q.empty() || keepalive; }
-
-    entity_addr_t& get_peer_addr() { return peer_addr; }
-
-    void set_peer_addr(const entity_addr_t& a) {
-      if (&peer_addr != &a)  // shut up valgrind
-        peer_addr = a;
-      connection_state->set_peer_addr(a);
-    }
-    void set_peer_type(int t) {
-      peer_type = t;
-      connection_state->set_peer_type(t);
-    }
-
-    void register_pipe();
-    void unregister_pipe();
-    void join() {
-      if (writer_thread.is_started())
-        writer_thread.join();
-      if (reader_thread.is_started())
-        reader_thread.join();
-    }
-    void stop();
-
-    void send(Message *m) {
-      pipe_lock.Lock();
-      _send(m);
-      pipe_lock.Unlock();
-    }
-    void _send(Message *m) {
-      out_q[m->get_priority()].push_back(m);
-      cond.Signal();
-    }
-    void _send_keepalive() {
-      keepalive = true;
-      cond.Signal();
-    }
-    Message *_get_next_outgoing() {
-      Message *m = 0;
-      while (!m && !out_q.empty()) {
-        map<int, list<Message*> >::reverse_iterator p = out_q.rbegin();
-        if (!p->second.empty()) {
-          m = p->second.front();
-          p->second.pop_front();
-        }
-        if (p->second.empty())
-          out_q.erase(p->first);
-      }
-      return m;
-    }
-
-    /* Remove all messages from the sent queue. Add those with seq > max_acked
-     * to the highest priority outgoing queue. */
-    void requeue_sent(uint64_t max_acked=0);
-    void discard_queue();
-
-    void shutdown_socket() {
-      if (sd >= 0)
-        ::shutdown(sd, SHUT_RDWR);
-    }
-  };
-
-  /**
-   * The DispatchQueue contains all the Pipes which have Messages
-   * they want to be dispatched, carefully organized by Message priority
-   * and permitted to deliver in a round-robin fashion.
-   * See SimpleMessenger::dispatch_entry for details.
-   */
-  struct DispatchQueue {
-    CephContext *cct;
-    SimpleMessenger *msgr;
-    Mutex lock;
-    Cond cond;
-    bool stop;
-
-    map<int, xlist<IncomingQueue *>* > queued_pipes;
-    map<int, xlist<IncomingQueue *>::iterator> queued_pipe_iters;
-    atomic_t qlen;
-    
-    enum { D_CONNECT = 1, D_BAD_REMOTE_RESET, D_BAD_RESET, D_NUM_CODES };
-    list<Connection*> connect_q;
-    list<Connection*> remote_reset_q;
-    list<Connection*> reset_q;
-
-    Pipe *local_pipe;
-    void local_delivery(Message *m, int priority) {
-      local_pipe->pipe_lock.Lock();
-      if ((unsigned long)m > 10)
-	m->set_connection(local_pipe->connection_state->get());
-      local_pipe->queue_received(m, priority);
-      local_pipe->pipe_lock.Unlock();
-    }
-
-    int get_queue_len() {
-      return qlen.read();
-    }
-    
-    void queue_connect(Connection *con) {
-      lock.Lock();
-      connect_q.push_back(con);
-      lock.Unlock();
-      local_delivery((Message*)D_CONNECT, CEPH_MSG_PRIO_HIGHEST);
-    }
-    void queue_remote_reset(Connection *con) {
-      lock.Lock();
-      remote_reset_q.push_back(con);
-      lock.Unlock();
-      local_delivery((Message*)D_BAD_REMOTE_RESET, CEPH_MSG_PRIO_HIGHEST);
-    }
-    void queue_reset(Connection *con) {
-      lock.Lock();
-      reset_q.push_back(con);
-      lock.Unlock();
-      local_delivery((Message*)D_BAD_RESET, CEPH_MSG_PRIO_HIGHEST);
-    }
-
-    void entry();
-
-    DispatchQueue(CephContext *cct, SimpleMessenger *msgr)
-      : cct(cct), msgr(msgr),
-	lock("SimpleMessenger::DispatchQeueu::lock"), 
-	stop(false),
-	qlen(0),
-	local_pipe(NULL)
-    {}
-    ~DispatchQueue() {
-      for (map< int, xlist<IncomingQueue *>* >::iterator i = queued_pipes.begin();
-	   i != queued_pipes.end();
-	   ++i) {
-	i->second->clear();
-	delete i->second;
-      }
-    }
-  } dispatch_queue;
+private:
 
   /**
    * A thread used to tear down Pipes when they're complete.
@@ -796,19 +413,6 @@ private:
     }
   } reaper_thread;
 
-  /**
-   * The DispatchThread runs dispatch_entry to empty out the dispatch_queue.
-   */
-  class DispatchThread : public Thread {
-    SimpleMessenger *msgr;
-  public:
-    DispatchThread(SimpleMessenger *_messenger) : msgr(_messenger) {}
-    void *entry() {
-      msgr->dispatch_entry();
-      return 0;
-    }
-  } dispatch_thread;
-
   /**
    * @} // Inner classes
    */
@@ -817,6 +421,7 @@ private:
    * @defgroup Utility functions
    * @{
    */
+
   /**
    * Create a Pipe associated with the given entity (of the given type).
    * Initiate the connection. (This function returning does not guarantee
@@ -874,6 +479,11 @@ private:
   /// true, specifying we haven't learned our addr; set false when we find it.
   // maybe this should be protected by the lock?
   bool need_addr;
+
+public:
+  bool get_need_addr() const { return need_addr; }
+
+private:
   /**
    *  false; set to true if the SimpleMessenger bound to a specific address;
    *  and set false again by Accepter::stop(). This isn't lock-protected
@@ -885,9 +495,6 @@ private:
   /// lock to protect the global_seq
   pthread_spinlock_t global_seq_lock;
 
-  /// flag set true when all the threads need to shut down
-  bool destination_stopped;
-
   /// hash map of addresses to Pipes
   hash_map<entity_addr_t, Pipe*> rank_pipe;
   /// a set of all the Pipes we have which are somehow active
@@ -897,6 +504,9 @@ private:
 
   /// internal cluster protocol version, if any, for talking to entities of the same type.
   int cluster_protocol;
+
+  /// lock protecting policy
+  Mutex policy_lock;
   /// the default Policy we use for Pipes
   Policy default_policy;
   /// map specifying different Policies for specific peer types
@@ -911,11 +521,15 @@ private:
   /// This Cond is slept on by wait() and signaled by dispatch_entry()
   Cond  wait_cond;
 
-  int timeout;
-
-  SimpleMessenger *msgr; //hack to make dout macro work, will fix
+  friend class Pipe;
 
 public:
+
+  int timeout;
+
+  /// con used for sending messages to ourselves
+  Connection *local_connection;
+
   /**
    * @defgroup SimpleMessenger internals
    * @{
@@ -953,12 +567,11 @@ public:
   int get_proto_version(int peer_type, bool connect);
 
   /**
-   * Fill in the address and peer type for the local pipe, which
-   * is used for delivering messages back to ourself (whether actual Messages
-   * submitted by the endpoint, or interrupts we use to process things like
-   * dead Pipes in a fair order).
+   * Fill in the address and peer type for the local connection, which
+   * is used for delivering messages back to ourself.
    */
-  void init_local_pipe();  /**
+  void init_local_connection();
+  /**
    * Tell the SimpleMessenger its full IP address.
    *
    * This is used by Pipes when connecting to other endpoints, and
@@ -966,28 +579,31 @@ public:
    */
   void learned_addr(const entity_addr_t& peer_addr_for_me);
 
+  /**
+   * Tell the SimpleMessenger its address is no longer known
+   *
+   * This happens when we rebind to a new port.
+   */
+  void unlearn_addr();
+
   /**
    * Get the Policy associated with a type of peer.
    * @param t The peer type to get the default policy for.
    *
    * @return A const Policy reference.
    */
-  const Policy& get_policy(int t) {
+  Policy get_policy(int t) {
+    Mutex::Locker l(policy_lock);
     if (policy_map.count(t))
       return policy_map[t];
     else
       return default_policy;
   }
+  Policy get_default_policy() {
+    Mutex::Locker l(policy_lock);
+    return default_policy;
+  }
 
-  /**
-   * This function is used by the dispatch thread. It runs continuously
-   * until dispatch_queue.stop is set to true, choosing what order the Pipes
-   * get to deliver in, and sending out their chosen Message via the
-   * ms_deliver_* functions.
-   * It should really only by dispatch_thread calling this, in our
-   * current implementation.
-   */
-  void dispatch_entry();
   /**
    * Release memory accounting back to the dispatch throttler.
    *
