@@ -171,7 +171,7 @@ public:
     version_t last_requested;           // last object requested by primary
 
     /****/
-    IndexedLog() {}
+    IndexedLog() : last_requested(0) {}
 
     void claim_log(const pg_log_t& o) {
       log = o.log;
@@ -281,7 +281,8 @@ public:
     uint64_t zero_to;                // first non-zeroed byte of log.
     bool has_checksums;
 
-    OndiskLog() : tail(0), head(0), zero_to(0) {}
+    OndiskLog() : tail(0), head(0), zero_to(0),
+		  has_checksums(true) {}
 
     uint64_t length() { return head - tail; }
     bool trim_to(eversion_t v, ObjectStore::Transaction& t);
@@ -347,6 +348,7 @@ protected:
    * put_unlock() when done with the current pointer (_most common_).
    */  
   Mutex _lock;
+  Mutex _qlock;
   Cond _cond;
   atomic_t ref;
 
@@ -354,12 +356,18 @@ public:
   bool deleting;  // true while RemoveWQ should be chewing on us
 
   void lock(bool no_lockdep = false);
+  void lockq(bool no_lockdep = false) {
+    _qlock.Lock(no_lockdep);
+  }
   void unlock() {
     //generic_dout(0) << this << " " << info.pgid << " unlock" << dendl;
     assert(!dirty_info);
     assert(!dirty_log);
     _lock.Unlock();
   }
+  void unlockq() {
+    _qlock.Unlock();
+  }
 
   /* During handle_osd_map, the osd holds a write lock to the osdmap.
    * *_with_map_lock_held assume that the map_lock is already held */
@@ -395,7 +403,6 @@ public:
   }
 
 
-  list<OpRequestRef> op_waiters;
   list<OpRequestRef> op_queue;  // op queue
 
   bool dirty_info, dirty_log;
@@ -432,6 +439,8 @@ protected:
   int         role;    // 0 = primary, 1 = replica, -1=none.
   unsigned    state;   // PG_STATE_*
 
+  bool send_notify;    ///< true if we are non-primary and should notify the primary
+
 public:
   eversion_t  last_update_ondisk;    // last_update that has committed; ONLY DEFINED WHEN is_active()
   eversion_t  last_complete_ondisk;  // last_complete that has committed.
@@ -496,7 +505,7 @@ public:
     const char *state_name;
     utime_t enter_time;
     const char *get_state_name() { return state_name; }
-    NamedState() : enter_time(ceph_clock_now(g_ceph_context)) {}
+    NamedState() : state_name(0), enter_time(ceph_clock_now(g_ceph_context)) {}
     virtual ~NamedState() {}
   };
 
@@ -610,6 +619,8 @@ protected:
   BackfillInterval backfill_info;
   BackfillInterval peer_backfill_info;
   int backfill_target;
+  bool backfill_reserved;
+  bool backfill_reserving;
 
   friend class OSD;
 
@@ -623,6 +634,11 @@ protected:
 
   // pg waiters
   bool flushed;
+
+  // Ops waiting on backfill_pos to change
+  list<OpRequestRef> waiting_for_backfill_pos;
+
+  list<OpRequestRef> waiting_for_map;
   list<OpRequestRef>            waiting_for_active;
   list<OpRequestRef>            waiting_for_all_missing;
   map<hobject_t, list<OpRequestRef> > waiting_for_missing_object,
@@ -662,6 +678,8 @@ public:
   bool needs_recovery() const;
 
   void mark_clean();  ///< mark an active pg clean
+
+  bool _calc_past_interval_range(epoch_t *start, epoch_t *end);
   void generate_past_intervals();
   void trim_past_intervals();
   void build_prior(std::auto_ptr<PriorSet> &prior_set);
@@ -768,17 +786,99 @@ public:
 
 
   // -- scrub --
-  set<int> scrub_reserved_peers;
-  map<int,ScrubMap> scrub_received_maps;
-  bool finalizing_scrub;
-  bool scrub_block_writes;
-  bool scrub_active;
-  bool scrub_reserved, scrub_reserve_failed;
-  int scrub_waiting_on;
-  set<int> scrub_waiting_on_whom;
-  epoch_t scrub_epoch_start;
-  ScrubMap primary_scrubmap;
-  MOSDRepScrub *active_rep_scrub;
+  struct Scrubber {
+    Scrubber() :
+      reserved(false), reserve_failed(false),
+      epoch_start(0),
+      block_writes(false), active(false), waiting_on(0),
+      errors(0), fixed(0), active_rep_scrub(0),
+      finalizing(false), is_chunky(false), state(INACTIVE),
+      deep(false)
+    {
+    }
+
+    // metadata
+    set<int> reserved_peers;
+    bool reserved, reserve_failed;
+    epoch_t epoch_start;
+
+    // common to both scrubs
+    bool block_writes;
+    bool active;
+    int waiting_on;
+    set<int> waiting_on_whom;
+    int errors;
+    int fixed;
+    ScrubMap primary_scrubmap;
+    map<int,ScrubMap> received_maps;
+    MOSDRepScrub *active_rep_scrub;
+
+    // classic scrub
+    bool finalizing;
+
+    // chunky scrub
+    bool is_chunky;
+    hobject_t start, end;
+    eversion_t subset_last_update;
+
+    // chunky scrub state
+    enum State {
+      INACTIVE,
+      NEW_CHUNK,
+      WAIT_PUSHES,
+      WAIT_LAST_UPDATE,
+      BUILD_MAP,
+      WAIT_REPLICAS,
+      COMPARE_MAPS,
+      FINISH,
+    } state;
+
+    // deep scrub
+    bool deep;
+
+    static const char *state_string(const PG::Scrubber::State& state) {
+      const char *ret = NULL;
+      switch( state )
+      {
+        case INACTIVE: ret = "INACTIVE"; break;
+        case NEW_CHUNK: ret = "NEW_CHUNK"; break;
+        case WAIT_PUSHES: ret = "WAIT_PUSHES"; break;
+        case WAIT_LAST_UPDATE: ret = "WAIT_LAST_UPDATE"; break;
+        case BUILD_MAP: ret = "BUILD_MAP"; break;
+        case WAIT_REPLICAS: ret = "WAIT_REPLICAS"; break;
+        case COMPARE_MAPS: ret = "COMPARE_MAPS"; break;
+        case FINISH: ret = "FINISH"; break;
+      }
+      return ret;
+    }
+
+    bool is_chunky_scrub_active() const { return state != INACTIVE; }
+
+    // clear all state
+    void reset() {
+      finalizing = false;
+      block_writes = false;
+      active = false;
+      waiting_on = 0;
+      waiting_on_whom.clear();
+      if (active_rep_scrub) {
+        active_rep_scrub->put();
+        active_rep_scrub = NULL;
+      }
+      received_maps.clear();
+
+      state = PG::Scrubber::INACTIVE;
+      start = hobject_t();
+      end = hobject_t();
+      subset_last_update = eversion_t();
+      errors = 0;
+      fixed = 0;
+      deep = false;
+    }
+
+  } scrubber;
+
+  int active_pushes;
 
   void repair_object(const hobject_t& soid, ScrubMap::object *po, int bad_peer, int ok_peer);
   bool _compare_scrub_objects(ScrubMap::object &auth,
@@ -790,14 +890,24 @@ public:
 			  map<hobject_t, int> &authoritative,
 			  ostream &errorstream);
   void scrub();
+  void classic_scrub();
+  void chunky_scrub();
+  void scrub_compare_maps();
   void scrub_finalize();
+  void scrub_finish();
   void scrub_clear_state();
   bool scrub_gather_replica_maps();
-  void _scan_list(ScrubMap &map, vector<hobject_t> &ls);
-  void _request_scrub_map(int replica, eversion_t version);
+  void _scan_list(ScrubMap &map, vector<hobject_t> &ls, bool deep);
+  void _request_scrub_map_classic(int replica, eversion_t version);
+  void _request_scrub_map(int replica, eversion_t version,
+                          hobject_t start, hobject_t end, bool deep);
+  int build_scrub_map_chunk(ScrubMap &map,
+                            hobject_t start, hobject_t end, bool deep);
   void build_scrub_map(ScrubMap &map);
   void build_inc_scrub_map(ScrubMap &map, eversion_t v);
-  virtual int _scrub(ScrubMap &map, int& errors, int& fixed) { return 0; }
+  virtual void _scrub(ScrubMap &map) { }
+  virtual void _scrub_clear_state() { }
+  virtual void _scrub_finish() { }
   virtual coll_t get_temp_coll() = 0;
   virtual bool have_temp_coll() = 0;
   void clear_scrub_reserved();
@@ -932,48 +1042,24 @@ public:
       *out << "Activate from " << query_epoch << std::endl;
     }
   };
-  struct Initialize : boost::statechart::event< Initialize > {
-    Initialize() : boost::statechart::event< Initialize >() {}
-    void print(std::ostream *out) const {
-      *out << "Initialize" << std::endl;
-    }
-  };
-  struct Load : boost::statechart::event< Load > {
-    Load() : boost::statechart::event< Load >() {}
-    void print(std::ostream *out) const {
-      *out << "Load" << std::endl;
-    }
-  };
-  struct GotInfo : boost::statechart::event< GotInfo > {
-    GotInfo() : boost::statechart::event< GotInfo >() {}
-    void print(std::ostream *out) const {
-      *out << "GotInfo" << std::endl;
-    }
-  };
-  struct NeedUpThru : boost::statechart::event< NeedUpThru > {
-    NeedUpThru() : boost::statechart::event< NeedUpThru >() {};
-    void print(std::ostream *out) const {
-      *out << "NeedUpThru" << std::endl;
-    }
-  };
-  struct CheckRepops : boost::statechart::event< CheckRepops > {
-    CheckRepops() : boost::statechart::event< CheckRepops >() {};
-    void print(std::ostream *out) const {
-      *out << "CheckRepops" << std::endl;
-    }
-  };
-  struct NullEvt : boost::statechart::event< NullEvt > {
-    NullEvt() : boost::statechart::event< NullEvt >() {};
-    void print(std::ostream *out) const {
-      *out << "NullEvt" << std::endl;
-    }
-  };
-  struct FlushedEvt : boost::statechart::event< FlushedEvt > {
-    FlushedEvt() : boost::statechart::event< FlushedEvt >() {};
-    void print(std::ostream *out) const {
-      *out << "FlushedEvt" << std::endl;
-    }
+#define TrivialEvent(T) struct T : boost::statechart::event< T > { \
+    T() : boost::statechart::event< T >() {}			   \
+    void print(std::ostream *out) const {			   \
+      *out << #T << std::endl;					   \
+    }								   \
   };
+  TrivialEvent(Initialize)
+  TrivialEvent(Load)
+  TrivialEvent(GotInfo)
+  TrivialEvent(NeedUpThru)
+  TrivialEvent(CheckRepops)
+  TrivialEvent(NullEvt)
+  TrivialEvent(FlushedEvt)
+  TrivialEvent(Backfilled)
+  TrivialEvent(LocalBackfillReserved)
+  TrivialEvent(RemoteBackfillReserved)
+  TrivialEvent(RemoteReservationRejected)
+  TrivialEvent(RequestBackfill)
 
   /* Encapsulates PG recovery process */
   class RecoveryState {
@@ -1073,11 +1159,12 @@ public:
 
       typedef boost::mpl::list <
 	boost::statechart::transition< Initialize, Reset >,
-	boost::statechart::transition< Load, Reset >,
+	boost::statechart::custom_reaction< Load >,
 	boost::statechart::custom_reaction< NullEvt >,
 	boost::statechart::transition< boost::statechart::event_base, Crashed >
 	> reactions;
 
+      boost::statechart::result react(const Load&);
       boost::statechart::result react(const MNotifyRec&);
       boost::statechart::result react(const MInfoRec&);
       boost::statechart::result react(const MLogRec&);
@@ -1164,8 +1251,7 @@ public:
       typedef boost::mpl::list <
 	boost::statechart::custom_reaction< ActMap >,
 	boost::statechart::custom_reaction< MNotifyRec >,
-	boost::statechart::transition< NeedActingChange, WaitActingChange >,
-	boost::statechart::transition< IsIncomplete, Incomplete >
+	boost::statechart::transition< NeedActingChange, WaitActingChange >
 	> reactions;
       boost::statechart::result react(const ActMap&);
       boost::statechart::result react(const MNotifyRec&);
@@ -1189,12 +1275,6 @@ public:
       void exit();
     };
 
-    struct Incomplete : boost::statechart::state< Incomplete, Primary>,
-			NamedState {
-      Incomplete(my_context ctx);
-      void exit();
-    };
-    
     struct GetInfo;
     struct Active;
 
@@ -1214,7 +1294,8 @@ public:
       boost::statechart::result react(const AdvMap &advmap);
     };
 
-    struct Active : boost::statechart::state< Active, Primary >, NamedState {
+    struct NotBackfilling;
+    struct Active : boost::statechart::state< Active, Primary, NotBackfilling >, NamedState {
       Active(my_context ctx);
       void exit();
 
@@ -1225,7 +1306,8 @@ public:
 	boost::statechart::custom_reaction< MInfoRec >,
 	boost::statechart::custom_reaction< MNotifyRec >,
 	boost::statechart::custom_reaction< MLogRec >,
-	boost::statechart::custom_reaction< RecoveryComplete >
+	boost::statechart::custom_reaction< RecoveryComplete >,
+	boost::statechart::custom_reaction< Backfilled >
 	> reactions;
       boost::statechart::result react(const QueryState& q);
       boost::statechart::result react(const ActMap&);
@@ -1233,10 +1315,49 @@ public:
       boost::statechart::result react(const MInfoRec& infoevt);
       boost::statechart::result react(const MNotifyRec& notevt);
       boost::statechart::result react(const MLogRec& logevt);
+      boost::statechart::result react(const Backfilled&) {
+	return discard_event();
+      }
       boost::statechart::result react(const RecoveryComplete&);
     };
 
-    struct ReplicaActive : boost::statechart::state< ReplicaActive, Started >, NamedState {
+    struct Backfilling : boost::statechart::state< Backfilling, Active >, NamedState {
+      typedef boost::mpl::list<
+	boost::statechart::transition< Backfilled, NotBackfilling >
+	> reactions;
+      Backfilling(my_context ctx);
+      void exit();
+    };
+
+    struct WaitRemoteBackfillReserved : boost::statechart::state< WaitRemoteBackfillReserved, Active >, NamedState {
+      typedef boost::mpl::list<
+	boost::statechart::custom_reaction< RemoteBackfillReserved >,
+	boost::statechart::custom_reaction< RemoteReservationRejected >
+	> reactions;
+      WaitRemoteBackfillReserved(my_context ctx);
+      void exit();
+      boost::statechart::result react(const RemoteBackfillReserved& evt);
+      boost::statechart::result react(const RemoteReservationRejected& evt);
+    };
+
+    struct WaitLocalBackfillReserved : boost::statechart::state< WaitLocalBackfillReserved, Active >, NamedState {
+      typedef boost::mpl::list<
+	boost::statechart::transition< LocalBackfillReserved, WaitRemoteBackfillReserved >
+	> reactions;
+      WaitLocalBackfillReserved(my_context ctx);
+      void exit();
+    };
+
+    struct NotBackfilling : boost::statechart::state< NotBackfilling, Active>, NamedState {
+      typedef boost::mpl::list<
+	boost::statechart::transition< RequestBackfill, WaitLocalBackfillReserved>
+	> reactions;
+      NotBackfilling(my_context ctx);
+      void exit();
+    };
+
+    struct RepNotBackfilling;
+    struct ReplicaActive : boost::statechart::state< ReplicaActive, Started, RepNotBackfilling >, NamedState {
       ReplicaActive(my_context ctx);
       void exit();
 
@@ -1256,6 +1377,33 @@ public:
       boost::statechart::result react(const Activate&);
     };
 
+    struct RepBackfilling : boost::statechart::state< RepBackfilling, ReplicaActive >, NamedState {
+      typedef boost::mpl::list<
+	boost::statechart::transition< Backfilled, RepNotBackfilling >
+	> reactions;
+      RepBackfilling(my_context ctx);
+      void exit();
+    };
+
+    struct RepWaitBackfillReserved : boost::statechart::state< RepWaitBackfillReserved, ReplicaActive >, NamedState {
+      typedef boost::mpl::list<
+	boost::statechart::custom_reaction< RemoteBackfillReserved >,
+	boost::statechart::custom_reaction< RemoteReservationRejected >
+	> reactions;
+      RepWaitBackfillReserved(my_context ctx);
+      void exit();
+      boost::statechart::result react(const RemoteBackfillReserved &evt);
+      boost::statechart::result react(const RemoteReservationRejected &evt);
+    };
+
+    struct RepNotBackfilling : boost::statechart::state< RepNotBackfilling, ReplicaActive>, NamedState {
+      typedef boost::mpl::list<
+	boost::statechart::transition< RequestBackfill, RepWaitBackfillReserved >
+	> reactions;
+      RepNotBackfilling(my_context ctx);
+      void exit();
+    };
+
     struct Stray : boost::statechart::state< Stray, Started >, NamedState {
       map<int, pair<pg_query_t, epoch_t> > pending_queries;
 
@@ -1307,8 +1455,11 @@ public:
       typedef boost::mpl::list <
 	boost::statechart::custom_reaction< QueryState >,
 	boost::statechart::custom_reaction< MLogRec >,
-	boost::statechart::custom_reaction< GotLog >
+	boost::statechart::custom_reaction< GotLog >,
+	boost::statechart::custom_reaction< AdvMap >,
+	boost::statechart::transition< IsIncomplete, Incomplete >
 	> reactions;
+      boost::statechart::result react(const AdvMap&);
       boost::statechart::result react(const QueryState& q);
       boost::statechart::result react(const MLogRec& logevt);
       boost::statechart::result react(const GotLog&);
@@ -1360,6 +1511,11 @@ public:
       boost::statechart::result react(const MLogRec& logrec);
     };
 
+    struct Incomplete : boost::statechart::state< Incomplete, Peering>, NamedState {
+      Incomplete(my_context ctx);
+      void exit();
+    };
+
 
     RecoveryMachine machine;
     PG *pg;
@@ -1389,7 +1545,7 @@ public:
 
  public:
   PG(OSDService *o, OSDMapRef curmap,
-     PGPool pool, pg_t p, const hobject_t& loid, const hobject_t& ioid);
+     const PGPool &pool, pg_t p, const hobject_t& loid, const hobject_t& ioid);
   virtual ~PG();
 
  private:
@@ -1417,6 +1573,7 @@ public:
   void state_clear(int m) { state &= ~m; }
 
   bool is_complete() const { return info.last_complete == info.last_update; }
+  bool should_send_notify() const { return send_notify; }
 
   int get_state() const { return state; }
   bool       is_active() const { return state_test(PG_STATE_ACTIVE); }
@@ -1425,7 +1582,6 @@ public:
   bool       is_replay() const { return state_test(PG_STATE_REPLAY); }
   bool       is_clean() const { return state_test(PG_STATE_CLEAN); }
   bool       is_degraded() const { return state_test(PG_STATE_DEGRADED); }
-  bool       is_stray() const { return state_test(PG_STATE_STRAY); }
 
   bool       is_scrubbing() const { return state_test(PG_STATE_SCRUBBING); }
 
@@ -1501,6 +1657,9 @@ public:
     return e <= get_osdmap()->get_epoch();
   }
 
+  bool op_has_sufficient_caps(OpRequestRef op);
+
+
   // recovery bits
   void take_waiters();
   void queue_peering_event(CephPeeringEvtRef evt);
@@ -1564,9 +1723,7 @@ WRITE_CLASS_ENCODER(PG::OndiskLog)
 
 ostream& operator<<(ostream& out, const PG& pg);
 
-namespace boost {
-  void intrusive_ptr_add_ref(PG *pg);
-  void intrusive_ptr_release(PG *pg);
-};
+void intrusive_ptr_add_ref(PG *pg);
+void intrusive_ptr_release(PG *pg);
 
 #endif
