@@ -280,6 +280,24 @@ bool coll_t::is_pg(pg_t& pgid, snapid_t& snap) const
   return true;
 }
 
+bool coll_t::is_removal(uint64_t *seq, pg_t *pgid) const
+{
+  if (str.substr(0, 11) != string("FORREMOVAL_"))
+    return false;
+
+  stringstream ss(str.substr(11));
+  ss >> *seq;
+  char sep;
+  ss >> sep;
+  assert(sep == '_');
+  string pgid_str;
+  ss >> pgid_str;
+  if (!pgid->parse(pgid_str.c_str())) {
+    assert(0);
+    return false;
+  }
+  return true;
+}
 
 void coll_t::encode(bufferlist& bl) const
 {
@@ -380,8 +398,6 @@ std::string pg_state_string(int state)
     oss << "down+";
   if (state & PG_STATE_REPLAY)
     oss << "replay+";
-  if (state & PG_STATE_STRAY)
-    oss << "stray+";
   if (state & PG_STATE_SPLITTING)
     oss << "splitting+";
   if (state & PG_STATE_DEGRADED)
@@ -390,6 +406,8 @@ std::string pg_state_string(int state)
     oss << "remapped+";
   if (state & PG_STATE_SCRUBBING)
     oss << "scrubbing+";
+  if (state & PG_STATE_DEEP_SCRUB)
+    oss << "deep+";
   if (state & PG_STATE_SCRUBQ)
     oss << "scrubq+";
   if (state & PG_STATE_INCONSISTENT)
@@ -398,8 +416,13 @@ std::string pg_state_string(int state)
     oss << "peering+";
   if (state & PG_STATE_REPAIR)
     oss << "repair+";
+  if (state & PG_STATE_BACKFILL_WAIT &&
+      !(state &PG_STATE_BACKFILL))
+    oss << "wait_backfill+";
   if (state & PG_STATE_BACKFILL)
-    oss << "backfill+";
+    oss << "backfilling+";
+  if (state & PG_STATE_BACKFILL_TOOFULL)
+    oss << "backfill_toofull+";
   if (state & PG_STATE_INCOMPLETE)
     oss << "incomplete+";
   string ret(oss.str());
@@ -955,6 +978,8 @@ void pg_stat_t::dump(Formatter *f) const
   f->dump_unsigned("parent_split_bits", parent_split_bits);
   f->dump_stream("last_scrub") << last_scrub;
   f->dump_stream("last_scrub_stamp") << last_scrub_stamp;
+  f->dump_stream("last_deep_scrub") << last_deep_scrub;
+  f->dump_stream("last_deep_scrub_stamp") << last_deep_scrub_stamp;
   f->dump_unsigned("log_size", log_size);
   f->dump_unsigned("ondisk_log_size", ondisk_log_size);
   stats.dump(f);
@@ -970,7 +995,7 @@ void pg_stat_t::dump(Formatter *f) const
 
 void pg_stat_t::encode(bufferlist &bl) const
 {
-  ENCODE_START(9, 8, bl);
+  ENCODE_START(10, 8, bl);
   ::encode(version, bl);
   ::encode(reported, bl);
   ::encode(state, bl);
@@ -993,12 +1018,14 @@ void pg_stat_t::encode(bufferlist &bl) const
   ::encode(last_clean, bl);
   ::encode(last_unstale, bl);
   ::encode(mapping_epoch, bl);
+  ::encode(last_deep_scrub, bl);
+  ::encode(last_deep_scrub_stamp, bl);
   ENCODE_FINISH(bl);
 }
 
 void pg_stat_t::decode(bufferlist::iterator &bl)
 {
-  DECODE_START_LEGACY_COMPAT_LEN(9, 8, 8, bl);
+  DECODE_START_LEGACY_COMPAT_LEN(10, 8, 8, bl);
   ::decode(version, bl);
   ::decode(reported, bl);
   ::decode(state, bl);
@@ -1056,6 +1083,10 @@ void pg_stat_t::decode(bufferlist::iterator &bl)
       ::decode(last_clean, bl);
       ::decode(last_unstale, bl);
       ::decode(mapping_epoch, bl);
+      if (struct_v >= 10) {
+        ::decode(last_deep_scrub, bl);
+        ::decode(last_deep_scrub_stamp, bl);
+      }
     }
   }
   DECODE_FINISH(bl);
@@ -1083,6 +1114,8 @@ void pg_stat_t::generate_test_instances(list<pg_stat_t*>& o)
   a.parent_split_bits = 12;
   a.last_scrub = eversion_t(9, 10);
   a.last_scrub_stamp = utime_t(11, 12);
+  a.last_deep_scrub = eversion_t(13, 14);
+  a.last_deep_scrub_stamp = utime_t(15, 16);
   list<object_stat_collection_t*> l;
   object_stat_collection_t::generate_test_instances(l);
   a.stats = *l.back();
@@ -1161,7 +1194,7 @@ void pool_stat_t::generate_test_instances(list<pool_stat_t*>& o)
 
 void pg_history_t::encode(bufferlist &bl) const
 {
-  ENCODE_START(4, 4, bl);
+  ENCODE_START(5, 4, bl);
   ::encode(epoch_created, bl);
   ::encode(last_epoch_started, bl);
   ::encode(last_epoch_clean, bl);
@@ -1171,12 +1204,14 @@ void pg_history_t::encode(bufferlist &bl) const
   ::encode(same_primary_since, bl);
   ::encode(last_scrub, bl);
   ::encode(last_scrub_stamp, bl);
+  ::encode(last_deep_scrub, bl);
+  ::encode(last_deep_scrub_stamp, bl);
   ENCODE_FINISH(bl);
 }
 
 void pg_history_t::decode(bufferlist::iterator &bl)
 {
-  DECODE_START_LEGACY_COMPAT_LEN(4, 4, 4, bl);
+  DECODE_START_LEGACY_COMPAT_LEN(5, 4, 4, bl);
   ::decode(epoch_created, bl);
   ::decode(last_epoch_started, bl);
   if (struct_v >= 3)
@@ -1190,6 +1225,10 @@ void pg_history_t::decode(bufferlist::iterator &bl)
   if (struct_v >= 2) {
     ::decode(last_scrub, bl);
     ::decode(last_scrub_stamp, bl);
+    if (struct_v >= 5) {
+      ::decode(last_deep_scrub, bl);
+      ::decode(last_deep_scrub_stamp, bl);
+    }
   }
   DECODE_FINISH(bl);
 }
@@ -1205,6 +1244,8 @@ void pg_history_t::dump(Formatter *f) const
   f->dump_int("same_primary_since", same_primary_since);
   f->dump_stream("last_scrub") << last_scrub;
   f->dump_stream("last_scrub_stamp") << last_scrub_stamp;
+  f->dump_stream("last_deep_scrub") << last_deep_scrub;
+  f->dump_stream("last_deep_scrub_stamp") << last_deep_scrub_stamp;
 }
 
 void pg_history_t::generate_test_instances(list<pg_history_t*>& o)
@@ -1219,7 +1260,9 @@ void pg_history_t::generate_test_instances(list<pg_history_t*>& o)
   o.back()->same_interval_since = 6;
   o.back()->same_primary_since = 7;
   o.back()->last_scrub = eversion_t(8, 9);
-  o.back()->last_scrub_stamp = utime_t(10, 11);  
+  o.back()->last_scrub_stamp = utime_t(10, 11);
+  o.back()->last_deep_scrub = eversion_t(12, 13);
+  o.back()->last_deep_scrub_stamp = utime_t(14, 15);
 }
 
 
@@ -1302,12 +1345,54 @@ void pg_info_t::generate_test_instances(list<pg_info_t*>& o)
   o.back()->last_update = eversion_t(3, 4);
   o.back()->last_complete = eversion_t(5, 6);
   o.back()->log_tail = eversion_t(7, 8);
-  o.back()->last_backfill = hobject_t(object_t("objname"), "key", 123, 456);
+  o.back()->last_backfill = hobject_t(object_t("objname"), "key", 123, 456, -1);
   list<pg_stat_t*> s;
   pg_stat_t::generate_test_instances(s);
   o.back()->stats = *s.back();
 }
 
+// -- pg_notify_t --
+void pg_notify_t::encode(bufferlist &bl) const
+{
+  ENCODE_START(1, 1, bl);
+  ::encode(query_epoch, bl);
+  ::encode(epoch_sent, bl);
+  ::encode(info, bl);
+  ENCODE_FINISH(bl);
+}
+
+void pg_notify_t::decode(bufferlist::iterator &bl)
+{
+  DECODE_START(1, bl);
+  ::decode(query_epoch, bl);
+  ::decode(epoch_sent, bl);
+  ::decode(info, bl);
+  DECODE_FINISH(bl);
+}
+
+void pg_notify_t::dump(Formatter *f) const
+{
+  f->dump_stream("query_epoch") << query_epoch;
+  f->dump_stream("epoch_sent") << epoch_sent;
+  {
+    f->open_object_section("info");
+    info.dump(f);
+    f->close_section();
+  }
+}
+
+void pg_notify_t::generate_test_instances(list<pg_notify_t*>& o)
+{
+  o.push_back(new pg_notify_t(1,1,pg_info_t()));
+  o.push_back(new pg_notify_t(3,10,pg_info_t()));
+}
+
+ostream &operator<<(ostream &lhs, const pg_notify_t &notify)
+{
+  return lhs << "(query_epoch:" << notify.query_epoch
+	     << ", epoch_sent:" << notify.epoch_sent
+	     << ", info:" << notify.info << ")";
+}
 
 // -- pg_interval_t --
 
@@ -1437,10 +1522,43 @@ ostream& operator<<(ostream& out, const pg_interval_t& i)
 
 // -- pg_query_t --
 
+void pg_query_t::encode(bufferlist &bl, uint64_t features) const {
+  if (features & CEPH_FEATURE_QUERY_T) {
+    ENCODE_START(2, 2, bl);
+    ::encode(type, bl);
+    ::encode(since, bl);
+    history.encode(bl);
+    ::encode(epoch_sent, bl);
+    ENCODE_FINISH(bl);
+  } else {
+    ::encode(type, bl);
+    ::encode(since, bl);
+    history.encode(bl);
+  }
+}
+
+void pg_query_t::decode(bufferlist::iterator &bl) {
+  bufferlist::iterator bl2 = bl;
+  try {
+    DECODE_START(2, bl);
+    ::decode(type, bl);
+    ::decode(since, bl);
+    history.decode(bl);
+    ::decode(epoch_sent, bl);
+    DECODE_FINISH(bl);
+  } catch (...) {
+    bl = bl2;
+    ::decode(type, bl);
+    ::decode(since, bl);
+    history.decode(bl);
+  }
+}
+
 void pg_query_t::dump(Formatter *f) const
 {
   f->dump_string("type", get_type_name());
   f->dump_stream("since") << since;
+  f->dump_stream("epoch_sent") << epoch_sent;
   f->open_object_section("history");
   history.dump(f);
   f->close_section();
@@ -1450,10 +1568,10 @@ void pg_query_t::generate_test_instances(list<pg_query_t*>& o)
   o.push_back(new pg_query_t());
   list<pg_history_t*> h;
   pg_history_t::generate_test_instances(h);
-  o.push_back(new pg_query_t(pg_query_t::INFO, *h.back()));
-  o.push_back(new pg_query_t(pg_query_t::MISSING, *h.back()));
-  o.push_back(new pg_query_t(pg_query_t::LOG, eversion_t(4, 5), *h.back()));
-  o.push_back(new pg_query_t(pg_query_t::FULLLOG, *h.back()));
+  o.push_back(new pg_query_t(pg_query_t::INFO, *h.back(), 4));
+  o.push_back(new pg_query_t(pg_query_t::MISSING, *h.back(), 4));
+  o.push_back(new pg_query_t(pg_query_t::LOG, eversion_t(4, 5), *h.back(), 4));
+  o.push_back(new pg_query_t(pg_query_t::FULLLOG, *h.back(), 5));
 }
 
 
@@ -1461,7 +1579,7 @@ void pg_query_t::generate_test_instances(list<pg_query_t*>& o)
 
 void pg_log_entry_t::encode(bufferlist &bl) const
 {
-  ENCODE_START(4, 4, bl);
+  ENCODE_START(5, 4, bl);
   ::encode(op, bl);
   ::encode(soid, bl);
   ::encode(version, bl);
@@ -1475,7 +1593,7 @@ void pg_log_entry_t::encode(bufferlist &bl) const
 
 void pg_log_entry_t::decode(bufferlist::iterator &bl)
 {
-  DECODE_START_LEGACY_COMPAT_LEN(4, 4, 4, bl);
+  DECODE_START_LEGACY_COMPAT_LEN(5, 4, 4, bl);
   ::decode(op, bl);
   if (struct_v < 2) {
     sobject_t old_soid;
@@ -1494,6 +1612,8 @@ void pg_log_entry_t::decode(bufferlist::iterator &bl)
   ::decode(mtime, bl);
   if (op == CLONE)
     ::decode(snaps, bl);
+  if (struct_v < 5)
+    invalid_pool = true;
   DECODE_FINISH(bl);
 }
 
@@ -1510,7 +1630,7 @@ void pg_log_entry_t::dump(Formatter *f) const
 void pg_log_entry_t::generate_test_instances(list<pg_log_entry_t*>& o)
 {
   o.push_back(new pg_log_entry_t());
-  hobject_t oid(object_t("objname"), "key", 123, 456);
+  hobject_t oid(object_t("objname"), "key", 123, 456, 0);
   o.push_back(new pg_log_entry_t(MODIFY, oid, eversion_t(1,2), eversion_t(3,4),
 				 osd_reqid_t(entity_name_t::CLIENT(777), 8, 999), utime_t(8,9)));
 }
@@ -1526,16 +1646,16 @@ ostream& operator<<(ostream& out, const pg_log_entry_t& e)
 
 void pg_log_t::encode(bufferlist& bl) const
 {
-  ENCODE_START(3, 3, bl);
+  ENCODE_START(4, 3, bl);
   ::encode(head, bl);
   ::encode(tail, bl);
   ::encode(log, bl);
   ENCODE_FINISH(bl);
 }
  
-void pg_log_t::decode(bufferlist::iterator &bl)
+void pg_log_t::decode(bufferlist::iterator &bl, int64_t pool)
 {
-  DECODE_START_LEGACY_COMPAT_LEN(3, 3, 3, bl);
+  DECODE_START_LEGACY_COMPAT_LEN(4, 3, 3, bl);
   ::decode(head, bl);
   ::decode(tail, bl);
   if (struct_v < 2) {
@@ -1544,6 +1664,16 @@ void pg_log_t::decode(bufferlist::iterator &bl)
   }
   ::decode(log, bl);
   DECODE_FINISH(bl);
+
+  // handle hobject_t format change
+  if (struct_v < 4) {
+    for (list<pg_log_entry_t>::iterator i = log.begin();
+	 i != log.end();
+	 ++i) {
+      if (i->soid.pool == -1)
+	i->soid.pool = pool;
+    }
+  }
 }
 
 void pg_log_t::dump(Formatter *f) const
@@ -1640,17 +1770,35 @@ ostream& pg_log_t::print(ostream& out) const
 
 void pg_missing_t::encode(bufferlist &bl) const
 {
-  ENCODE_START(2, 2, bl);
+  ENCODE_START(3, 2, bl);
   ::encode(missing, bl);
   ENCODE_FINISH(bl);
 }
 
-void pg_missing_t::decode(bufferlist::iterator &bl)
+void pg_missing_t::decode(bufferlist::iterator &bl, int64_t pool)
 {
-  DECODE_START_LEGACY_COMPAT_LEN(2, 2, 2, bl);
+  DECODE_START_LEGACY_COMPAT_LEN(3, 2, 2, bl);
   ::decode(missing, bl);
   DECODE_FINISH(bl);
 
+  if (struct_v < 3) {
+    // Handle hobject_t upgrade
+    map<hobject_t, item> tmp;
+    for (map<hobject_t, item>::iterator i = missing.begin();
+	 i != missing.end();
+      ) {
+      if (i->first.pool == -1) {
+	hobject_t to_insert(i->first);
+	to_insert.pool = pool;
+	tmp[to_insert] = i->second;
+	missing.erase(i++);
+      } else {
+	++i;
+      }
+    }
+    missing.insert(tmp.begin(), tmp.end());
+  }
+
   for (map<hobject_t,item>::iterator it = missing.begin();
        it != missing.end();
        ++it)
@@ -1673,7 +1821,7 @@ void pg_missing_t::generate_test_instances(list<pg_missing_t*>& o)
 {
   o.push_back(new pg_missing_t);
   o.push_back(new pg_missing_t);
-  o.back()->add(hobject_t(object_t("foo"), "foo", 123, 456), eversion_t(5, 6), eversion_t(5, 1));
+  o.back()->add(hobject_t(object_t("foo"), "foo", 123, 456, 0), eversion_t(5, 6), eversion_t(5, 1));
 }
 
 ostream& operator<<(ostream& out, const pg_missing_t::item& i) 
@@ -1775,6 +1923,13 @@ void pg_missing_t::revise_need(hobject_t oid, eversion_t need)
   rmissing[need.version] = oid;
 }
 
+void pg_missing_t::revise_have(hobject_t oid, eversion_t have)
+{
+  if (missing.count(oid)) {
+    missing[oid].have = have;
+  }
+}
+
 void pg_missing_t::add(const hobject_t& oid, eversion_t need, eversion_t have)
 {
   missing[oid] = item(need, have);
@@ -2086,7 +2241,7 @@ ps_t object_info_t::legacy_object_locator_to_ps(const object_t &oid,
 
 void object_info_t::encode(bufferlist& bl) const
 {
-  ENCODE_START(9, 8, bl);
+  ENCODE_START(10, 8, bl);
   ::encode(soid, bl);
   ::encode(oloc, bl);
   ::encode(category, bl);
@@ -2110,18 +2265,18 @@ void object_info_t::encode(bufferlist& bl) const
 
 void object_info_t::decode(bufferlist::iterator& bl)
 {
-  DECODE_START_LEGACY_COMPAT_LEN(9, 8, 8, bl);
+  DECODE_START_LEGACY_COMPAT_LEN(10, 8, 8, bl);
   if (struct_v >= 2 && struct_v <= 5) {
     sobject_t obj;
     ::decode(obj, bl);
     ::decode(oloc, bl);
-    soid = hobject_t(obj.oid, oloc.key, obj.snap, 0);
+    soid = hobject_t(obj.oid, oloc.key, obj.snap, 0, 0);
     soid.hash = legacy_object_locator_to_ps(soid.oid, oloc);
   } else if (struct_v >= 6) {
     ::decode(soid, bl);
     ::decode(oloc, bl);
     if (struct_v == 6) {
-      hobject_t hoid(soid.oid, oloc.key, soid.snap, hoid.hash);
+      hobject_t hoid(soid.oid, oloc.key, soid.snap, soid.hash, 0);
       soid = hoid;
     }
   }
@@ -2151,6 +2306,8 @@ void object_info_t::decode(bufferlist::iterator& bl)
     ::decode(uses_tmap, bl);
   else
     uses_tmap = true;
+  if (struct_v < 10)
+    soid.pool = oloc.pool;
   DECODE_FINISH(bl);
 }
 
@@ -2275,7 +2432,7 @@ void ObjectRecoveryProgress::dump(Formatter *f) const
 
 void ObjectRecoveryInfo::encode(bufferlist &bl) const
 {
-  ENCODE_START(1, 1, bl);
+  ENCODE_START(2, 1, bl);
   ::encode(soid, bl);
   ::encode(version, bl);
   ::encode(size, bl);
@@ -2286,9 +2443,10 @@ void ObjectRecoveryInfo::encode(bufferlist &bl) const
   ENCODE_FINISH(bl);
 }
 
-void ObjectRecoveryInfo::decode(bufferlist::iterator &bl)
+void ObjectRecoveryInfo::decode(bufferlist::iterator &bl,
+				int64_t pool)
 {
-  DECODE_START(1, bl);
+  DECODE_START(2, bl);
   ::decode(soid, bl);
   ::decode(version, bl);
   ::decode(size, bl);
@@ -2297,6 +2455,21 @@ void ObjectRecoveryInfo::decode(bufferlist::iterator &bl)
   ::decode(copy_subset, bl);
   ::decode(clone_subset, bl);
   DECODE_FINISH(bl);
+
+  if (struct_v < 2) {
+    if (soid.pool == -1)
+      soid.pool = pool;
+    map<hobject_t, interval_set<uint64_t> > tmp;
+    tmp.swap(clone_subset);
+    for (map<hobject_t, interval_set<uint64_t> >::iterator i = tmp.begin();
+	 i != tmp.end();
+	 ++i) {
+      hobject_t first(i->first);
+      if (first.pool == -1)
+	first.pool = pool;
+      clone_subset[first].swap(i->second);
+    }
+  }
 }
 
 void ObjectRecoveryInfo::generate_test_instances(
@@ -2367,7 +2540,7 @@ void ScrubMap::merge_incr(const ScrubMap &l)
 
 void ScrubMap::encode(bufferlist& bl) const
 {
-  ENCODE_START(2, 2, bl);
+  ENCODE_START(3, 2, bl);
   ::encode(objects, bl);
   ::encode(attrs, bl);
   ::encode(logbl, bl);
@@ -2376,15 +2549,29 @@ void ScrubMap::encode(bufferlist& bl) const
   ENCODE_FINISH(bl);
 }
 
-void ScrubMap::decode(bufferlist::iterator& bl)
+void ScrubMap::decode(bufferlist::iterator& bl, int64_t pool)
 {
-  DECODE_START_LEGACY_COMPAT_LEN(2, 2, 2, bl);
+  DECODE_START_LEGACY_COMPAT_LEN(3, 2, 2, bl);
   ::decode(objects, bl);
   ::decode(attrs, bl);
   ::decode(logbl, bl);
   ::decode(valid_through, bl);
   ::decode(incr_since, bl);
   DECODE_FINISH(bl);
+
+  // handle hobject_t upgrade
+  if (struct_v < 3) {
+    map<hobject_t, object> tmp;
+    tmp.swap(objects);
+    for (map<hobject_t, object>::iterator i = tmp.begin();
+	 i != tmp.end();
+	 ++i) {
+      hobject_t first(i->first);
+      if (first.pool == -1)
+	first.pool = pool;
+      objects[first] = i->second;
+    }
+  }
 }
 
 void ScrubMap::dump(Formatter *f) const
@@ -2422,28 +2609,38 @@ void ScrubMap::generate_test_instances(list<ScrubMap*>& o)
   o.back()->attrs["bar"] = buffer::copy("barval", 6);
   list<object*> obj;
   object::generate_test_instances(obj);
-  o.back()->objects[hobject_t(object_t("foo"), "fookey", 123, 456)] = *obj.back();
+  o.back()->objects[hobject_t(object_t("foo"), "fookey", 123, 456, 0)] = *obj.back();
   obj.pop_back();
-  o.back()->objects[hobject_t(object_t("bar"), string(), 123, 456)] = *obj.back();
+  o.back()->objects[hobject_t(object_t("bar"), string(), 123, 456, 0)] = *obj.back();
 }
 
 // -- ScrubMap::object --
 
 void ScrubMap::object::encode(bufferlist& bl) const
 {
-  ENCODE_START(2, 2, bl);
+  ENCODE_START(3, 2, bl);
   ::encode(size, bl);
   ::encode(negative, bl);
   ::encode(attrs, bl);
+  ::encode(digest, bl);
+  ::encode(digest_present, bl);
   ENCODE_FINISH(bl);
 }
 
 void ScrubMap::object::decode(bufferlist::iterator& bl)
 {
-  DECODE_START_LEGACY_COMPAT_LEN(2, 2, 2, bl);
+  DECODE_START_LEGACY_COMPAT_LEN(3, 2, 2, bl);
   ::decode(size, bl);
   ::decode(negative, bl);
   ::decode(attrs, bl);
+  if (struct_v >= 3) {
+    ::decode(digest, bl);
+    ::decode(digest_present, bl);
+  }
+  else {
+    digest = 0;
+    digest_present = false;
+  }
   DECODE_FINISH(bl);
 }
 
