@@ -551,7 +551,7 @@ inline ostream& operator<<(ostream& out, const osd_stat_t& s) {
 #define PG_STATE_CLEAN        (1<<2)  // peers are complete, clean of stray replicas.
 #define PG_STATE_DOWN         (1<<4)  // a needed replica is down, PG offline
 #define PG_STATE_REPLAY       (1<<5)  // crashed, waiting for replay
-#define PG_STATE_STRAY        (1<<6)  // i must notify the primary i exist.
+//#define PG_STATE_STRAY      (1<<6)  // i must notify the primary i exist.
 #define PG_STATE_SPLITTING    (1<<7)  // i am splitting
 #define PG_STATE_SCRUBBING    (1<<8)  // scrubbing
 #define PG_STATE_SCRUBQ       (1<<9)  // queued for scrub
@@ -560,10 +560,13 @@ inline ostream& operator<<(ostream& out, const osd_stat_t& s) {
 #define PG_STATE_PEERING      (1<<12) // pg is (re)peering
 #define PG_STATE_REPAIR       (1<<13) // pg should repair on next scrub
 #define PG_STATE_RECOVERING   (1<<14) // pg is recovering/migrating objects
-#define PG_STATE_BACKFILL     (1<<15) // [active] backfilling pg content
+#define PG_STATE_BACKFILL_WAIT     (1<<15) // [active] reserving backfill
 #define PG_STATE_INCOMPLETE   (1<<16) // incomplete content, peering failed.
 #define PG_STATE_STALE        (1<<17) // our state for this pg is stale, unknown.
 #define PG_STATE_REMAPPED     (1<<18) // pg is explicitly remapped to different OSDs than CRUSH
+#define PG_STATE_DEEP_SCRUB   (1<<19) // deep scrub: check CRC32 on files
+#define PG_STATE_BACKFILL  (1<<20) // [active] backfilling pg content
+#define PG_STATE_BACKFILL_TOOFULL (1<<21) // backfill can't proceed: too full
 
 std::string pg_state_string(int state);
 
@@ -862,7 +865,9 @@ struct pg_stat_t {
   __u32 parent_split_bits;
 
   eversion_t last_scrub;
+  eversion_t last_deep_scrub;
   utime_t last_scrub_stamp;
+  utime_t last_deep_scrub_stamp;
 
   object_stat_collection_t stats;
 
@@ -951,7 +956,9 @@ struct pg_history_t {
   epoch_t same_primary_since;  // same primary at least back through this epoch.
 
   eversion_t last_scrub;
+  eversion_t last_deep_scrub;
   utime_t last_scrub_stamp;
+  utime_t last_deep_scrub_stamp;
 
   pg_history_t()
     : epoch_created(0),
@@ -973,8 +980,8 @@ struct pg_history_t {
       last_epoch_clean = other.last_epoch_clean;
       modified = true;
     }
-    if (last_epoch_split < other.last_epoch_started) {
-      last_epoch_split = other.last_epoch_started; 
+    if (last_epoch_split < other.last_epoch_split) {
+      last_epoch_split = other.last_epoch_split; 
       modified = true;
     }
     if (other.last_scrub > last_scrub) {
@@ -985,6 +992,14 @@ struct pg_history_t {
       last_scrub_stamp = other.last_scrub_stamp;
       modified = true;
     }
+    if (other.last_deep_scrub > last_deep_scrub) {
+      last_deep_scrub = other.last_deep_scrub;
+      modified = true;
+    }
+    if (other.last_deep_scrub_stamp > last_deep_scrub_stamp) {
+      last_deep_scrub_stamp = other.last_deep_scrub_stamp;
+      modified = true;
+    }
     return modified;
   }
 
@@ -1085,7 +1100,7 @@ struct pg_notify_t {
   static void generate_test_instances(list<pg_notify_t*> &o);
 };
 WRITE_CLASS_ENCODER(pg_notify_t)
-ostream &operator<<(ostream &lhs, const pg_notify_t notify);
+ostream &operator<<(ostream &lhs, const pg_notify_t &notify);
 
 
 /**
@@ -1155,7 +1170,7 @@ struct pg_query_t {
   pg_history_t history;
   epoch_t epoch_sent;
 
-  pg_query_t() : type(-1) {}
+  pg_query_t() : type(-1), epoch_sent(0) {}
   pg_query_t(int t, const pg_history_t& h,
 	     epoch_t epoch_sent)
     : type(t), history(h),
@@ -1236,7 +1251,7 @@ struct pg_log_entry_t {
   uint64_t offset;   // [soft state] my offset on disk
       
   pg_log_entry_t()
-    : op(0), invalid_hash(false), offset(0) {}
+    : op(0), invalid_hash(false), invalid_pool(false), offset(0) {}
   pg_log_entry_t(int _op, const hobject_t& _soid, 
 		 const eversion_t& v, const eversion_t& pv,
 		 const osd_reqid_t& rid, const utime_t& mt)
@@ -1777,8 +1792,10 @@ struct ScrubMap {
     uint64_t size;
     bool negative;
     map<string,bufferptr> attrs;
+    __u32 digest;
+    bool digest_present;
 
-    object(): size(0), negative(false) {}
+    object(): size(0), negative(false), digest(0), digest_present(false) {}
 
     void encode(bufferlist& bl) const;
     void decode(bufferlist::iterator& bl);
