@@ -424,6 +424,22 @@ struct ObjectOperation {
     }
   }
 
+  void omap_cmp(const std::map<std::string, pair<bufferlist, int> > &assertions,
+		int *prval) {
+    OSDOp &op = add_op(CEPH_OSD_OP_OMAP_CMP);
+    bufferlist bl;
+    ::encode(assertions, bl);
+    op.op.extent.offset = 0;
+    op.op.extent.length = bl.length();
+    op.indata.claim_append(bl);
+    if (prval) {
+      unsigned p = ops.size() - 1;
+      out_rval[p] = prval;
+      out_bl[p] = NULL;
+      out_handler[p] = NULL;
+    }
+  }
+
   void omap_get_header(bufferlist *bl, int *prval) {
     add_op(CEPH_OSD_OP_OMAPGETHEADER);
     unsigned p = ops.size() - 1;
@@ -523,7 +539,7 @@ class Objecter {
   bool keep_balanced_budget;
   bool honor_osdmap_full;
 
-  void maybe_request_map(epoch_t epoch=0);
+  void maybe_request_map();
 
   version_t last_seen_osdmap_version;
   version_t last_seen_pgmap_version;
@@ -547,7 +563,7 @@ class Objecter {
     Objecter *m_objecter;
   public:
     RequestStateHook(Objecter *objecter);
-    bool call(std::string command, bufferlist& out);
+    bool call(std::string command, std::string args, bufferlist& out);
   };
 
   RequestStateHook *m_request_state_hook;
@@ -598,6 +614,13 @@ public:
 
     utime_t stamp;
 
+    bool precalc_pgid;
+
+    bool budgeted;
+
+    /// true if we should resend this message on failure
+    bool should_resend;
+
     Op(const object_t& o, const object_locator_t& ol, vector<OSDOp>& op,
        int f, Context *ac, Context *co, eversion_t *ov) :
       session(NULL), session_item(this), incarnation(0),
@@ -607,7 +630,9 @@ public:
       outbl(NULL),
       flags(f), priority(0), onack(ac), oncommit(co),
       tid(0), attempts(0),
-      paused(false), objver(ov), reply_epoch(NULL) {
+      paused(false), objver(ov), reply_epoch(NULL), precalc_pgid(false),
+      budgeted(false),
+      should_resend(true) {
       ops.swap(op);
       
       /* initialize out_* to match op vector */
@@ -708,7 +733,7 @@ public:
       list_context(lc), final_finish(finish), bl(b), objecter(ob), epoch(0) {}
     void finish(int r) {
       if (r >= 0) {
-        objecter->_list_reply(list_context, bl, final_finish, epoch);
+        objecter->_list_reply(list_context, r, bl, final_finish, epoch);
       } else {
         final_finish->finish(r);
         delete final_finish;
@@ -768,16 +793,19 @@ public:
     bufferlist *poutbl;
     eversion_t *pobjver;
 
-    bool registering, registered;
+    bool registered;
     Context *on_reg_ack, *on_reg_commit;
 
     OSDSession *session;
     xlist<LingerOp*>::item session_item;
 
+    tid_t register_tid;
+
     LingerOp() : linger_id(0), flags(0), poutbl(NULL), pobjver(NULL),
-		 registering(false), registered(false),
+		 registered(false),
 		 on_reg_ack(NULL), on_reg_commit(NULL),
-		 session(NULL), session_item(this) {}
+		 session(NULL), session_item(this),
+		 register_tid(0) {}
 
     // no copy!
     const LingerOp &operator=(const LingerOp& r);
@@ -852,6 +880,8 @@ public:
   map<epoch_t,list< pair<Context*, int> > > waiting_for_map;
 
   void send_op(Op *op);
+  void cancel_op(Op *op);
+  void finish_op(Op *op);
   bool is_pg_changed(vector<int>& a, vector<int>& b, bool any_change=false);
   enum recalc_op_target_result {
     RECALC_OP_TARGET_NO_ACTION = 0,
@@ -876,7 +906,7 @@ public:
   void reopen_session(OSDSession *session);
   void close_session(OSDSession *session);
   
-  void _list_reply(ListContext *list_context, bufferlist *bl, Context *final_finish,
+  void _list_reply(ListContext *list_context, int r, bufferlist *bl, Context *final_finish,
 		   epoch_t reply_epoch);
 
   void resend_mon_ops();
@@ -891,16 +921,21 @@ public:
   void throttle_op(Op *op, int op_size=0);
   void take_op_budget(Op *op) {
     int op_budget = calc_op_budget(op);
-    if (keep_balanced_budget)
+    if (keep_balanced_budget) {
       throttle_op(op, op_budget);
-    else
-      op_throttler.take(op_budget);
+    } else {
+      op_throttle_bytes.take(op_budget);
+      op_throttle_ops.take(1);
+    }
+    op->budgeted = true;
   }
   void put_op_budget(Op *op) {
+    assert(op->budgeted);
     int op_budget = calc_op_budget(op);
-    op_throttler.put(op_budget);
+    op_throttle_bytes.put(op_budget);
+    op_throttle_ops.put(1);
   }
-  Throttle op_throttler;
+  Throttle op_throttle_bytes, op_throttle_ops;
 
  public:
   Objecter(CephContext *cct_, Messenger *m, MonClient *mc,
@@ -917,7 +952,8 @@ public:
     logger(NULL), tick_event(NULL),
     m_request_state_hook(NULL),
     num_homeless_ops(0),
-    op_throttler(cct->_conf->objecter_inflight_op_bytes)
+    op_throttle_bytes(cct, "objecter_bytes", cct->_conf->objecter_inflight_op_bytes),
+    op_throttle_ops(cct, "objecter_ops", cct->_conf->objecter_inflight_ops)
   { }
   ~Objecter() {
     assert(!tick_event);
@@ -941,6 +977,10 @@ public:
   void set_honor_osdmap_full() { honor_osdmap_full = true; }
   void unset_honor_osdmap_full() { honor_osdmap_full = false; }
 
+  void scan_requests(bool skipped_map,
+		     map<tid_t, Op*>& need_resend,
+		     list<LingerOp*>& need_resend_linger);
+
   // messages
  public:
   void dispatch(Message *m);
@@ -950,7 +990,8 @@ public:
 
 private:
   // low-level
-  tid_t op_submit(Op *op, OSDSession *s = NULL);
+  tid_t op_submit(Op *op);
+  tid_t _op_submit(Op *op);
 
   // public interface
  public:
@@ -972,10 +1013,7 @@ private:
   int get_client_incarnation() const { return client_inc; }
   void set_client_incarnation(int inc) { client_inc = inc; }
 
-  void wait_for_new_map(Context *c, epoch_t epoch, int replyCode=0) {
-    maybe_request_map(epoch);
-    waiting_for_map[epoch].push_back(pair<Context *, int>(c, replyCode));
-  }
+  void wait_for_new_map(Context *c, epoch_t epoch, int err=0);
 
   /** Get the current set of global op flags */
   int get_global_op_flags() { return global_op_flags; }
@@ -1451,7 +1489,7 @@ public:
       C_GatherBuilder gcom(cct, oncommit);
       for (vector<ObjectExtent>::iterator p = extents.begin(); p != extents.end(); p++) {
 	bufferlist cur;
-	for (map<__u32,__u32>::iterator bit = p->buffer_extents.begin();
+	for (map<uint64_t, uint64_t>::iterator bit = p->buffer_extents.begin();
 	     bit != p->buffer_extents.end();
 	     bit++)
 	  bl.copy(bit->first, bit->second, cur);
