@@ -135,17 +135,6 @@ public:
   void set_ctx(CephContext *_cct) { cct = _cct; }
 };
 
-static inline void normalize_bucket_and_obj(rgw_bucket& src_bucket, string& src_obj, rgw_bucket& dst_bucket, string& dst_obj)
-{
-  if (src_obj.size()) {
-    dst_bucket = src_bucket;
-    dst_obj = src_obj;
-  } else {
-    dst_bucket = rgw_root_bucket;
-    dst_obj = src_bucket.name;
-  }
-}
-
 template <class T>
 class RGWCache  : public T
 {
@@ -167,6 +156,7 @@ class RGWCache  : public T
     return string(buf);
   }
 
+  void normalize_bucket_and_obj(rgw_bucket& src_bucket, string& src_obj, rgw_bucket& dst_bucket, string& dst_obj);
   string normal_name(rgw_obj& obj) {
     return normal_name(obj.bucket, obj.object);
   }
@@ -184,6 +174,7 @@ class RGWCache  : public T
 
   void finalize() {
     T::finalize_watch();
+    T::finalize();
   }
   int distribute(const string& normal_name, rgw_obj& obj, ObjectCacheInfo& obj_info, int op);
   int watch_cb(int opcode, uint64_t ver, bufferlist& bl);
@@ -206,18 +197,29 @@ public:
 
   int obj_stat(void *ctx, rgw_obj& obj, uint64_t *psize, time_t *pmtime, map<string, bufferlist> *attrs, bufferlist *first_chunk);
 
-  int delete_obj(void *ctx, rgw_obj& obj, bool sync);
+  int delete_obj(void *ctx, rgw_obj& obj);
 };
 
+template <class T>
+void RGWCache<T>::normalize_bucket_and_obj(rgw_bucket& src_bucket, string& src_obj, rgw_bucket& dst_bucket, string& dst_obj)
+{
+  if (src_obj.size()) {
+    dst_bucket = src_bucket;
+    dst_obj = src_obj;
+  } else {
+    dst_bucket = T::params.domain_root;
+    dst_obj = src_bucket.name;
+  }
+}
 
 template <class T>
-int RGWCache<T>::delete_obj(void *ctx, rgw_obj& obj, bool sync)
+int RGWCache<T>::delete_obj(void *ctx, rgw_obj& obj)
 {
   rgw_bucket bucket;
   string oid;
   normalize_bucket_and_obj(obj.bucket, obj.object, bucket, oid);
   if (bucket.name[0] != '.')
-    return T::delete_obj(ctx, obj, sync);
+    return T::delete_obj(ctx, obj);
 
   string name = normal_name(obj);
   cache.remove(name);
@@ -225,7 +227,7 @@ int RGWCache<T>::delete_obj(void *ctx, rgw_obj& obj, bool sync)
   ObjectCacheInfo info;
   distribute(name, obj, info, REMOVE_OBJ);
 
-  return T::delete_obj(ctx, obj, sync);
+  return T::delete_obj(ctx, obj);
 }
 
 template <class T>
