@@ -4,12 +4,13 @@
 #include "include/rados/librados.hpp"
 #include "include/Context.h"
 #include "rgw_common.h"
-#include "rgw_cls_api.h"
+#include "cls/rgw/cls_rgw_types.h"
 #include "rgw_log.h"
 
 class RGWWatcher;
 class SafeTimer;
 class ACLOwner;
+class RGWGC;
 
 static inline void prepend_bucket_marker(rgw_bucket& bucket, string& orig_oid, string& oid)
 {
@@ -29,6 +30,23 @@ static inline void get_obj_bucket_and_oid_key(rgw_obj& obj, rgw_bucket& bucket,
   prepend_bucket_marker(bucket, obj.key, key);
 }
 
+struct RGWUsageBatch {
+  map<utime_t, rgw_usage_log_entry> m;
+
+  void insert(utime_t& t, rgw_usage_log_entry& entry, bool *account) {
+    bool exists = m.find(t) != m.end();
+    *account = !exists;
+    m[t].aggregate(entry);
+  }
+};
+
+struct RGWUsageIter {
+  string read_iter;
+  uint32_t index;
+
+  RGWUsageIter() : index(0) {}
+};
+
 class RGWAccessListFilter {
 public:
   virtual ~RGWAccessListFilter() {}
@@ -101,7 +119,9 @@ struct RGWObjState {
   bool exists;
   uint64_t size;
   time_t mtime;
+  uint64_t epoch;
   bufferlist obj_tag;
+  bool fake_tag;
   RGWObjManifest manifest;
   bool has_manifest;
   string shadow_obj;
@@ -110,7 +130,7 @@ struct RGWObjState {
   bool prefetch_data;
 
   map<string, bufferlist> attrset;
-  RGWObjState() : is_atomic(false), has_attrs(0), exists(false), has_manifest(false), prefetch_data(false) {}
+  RGWObjState() : is_atomic(false), has_attrs(0), exists(false), epoch(0), fake_tag(false), has_manifest(false), prefetch_data(false) {}
 
   bool get_attr(string name, bufferlist& dest) {
     map<string, bufferlist>::iterator iter = attrset.find(name);
@@ -124,6 +144,8 @@ struct RGWObjState {
   void clear() {
     has_attrs = false;
     exists = false;
+    fake_tag = false;
+    epoch = 0;
     size = 0;
     mtime = 0;
     obj_tag.clear();
@@ -134,40 +156,24 @@ struct RGWObjState {
 };
 
 struct RGWRadosCtx {
+  RGWRados *store;
   map<rgw_obj, RGWObjState> objs_state;
-  int (*intent_cb)(void *user_ctx, rgw_obj& obj, RGWIntentEvent intent);
+  int (*intent_cb)(RGWRados *store, void *user_ctx, rgw_obj& obj, RGWIntentEvent intent);
   void *user_ctx;
-  RGWObjState *get_state(rgw_obj& obj) {
-    if (obj.object.size()) {
-      return &objs_state[obj];
-    } else {
-      rgw_obj new_obj(rgw_root_bucket, obj.bucket.name);
-      return &objs_state[new_obj];
-    }
-  }
-  void set_atomic(rgw_obj& obj) {
-    if (obj.object.size()) {
-      objs_state[obj].is_atomic = true;
-    } else {
-      rgw_obj new_obj(rgw_root_bucket, obj.bucket.name);
-      objs_state[new_obj].is_atomic = true;
-    }
-  }
-  void set_prefetch_data(rgw_obj& obj) {
-    if (obj.object.size()) {
-      objs_state[obj].prefetch_data = true;
-    } else {
-      rgw_obj new_obj(rgw_root_bucket, obj.bucket.name);
-      objs_state[new_obj].prefetch_data = true;
-    }
-  }
-  void set_intent_cb(int (*cb)(void *user_ctx, rgw_obj& obj, RGWIntentEvent intent)) {
+
+  RGWRadosCtx(RGWRados *_store) : store(_store) { }
+
+  RGWObjState *get_state(rgw_obj& obj);
+  void set_atomic(rgw_obj& obj);
+  void set_prefetch_data(rgw_obj& obj);
+
+  void set_intent_cb(int (*cb)(RGWRados *store, void *user_ctx, rgw_obj& obj, RGWIntentEvent intent)) {
     intent_cb = cb;
   }
 
-  int notify_intent(rgw_obj& obj, RGWIntentEvent intent) {
+  int notify_intent(RGWRados *store, rgw_obj& obj, RGWIntentEvent intent) {
     if (intent_cb) {
-      return intent_cb(user_ctx, obj, intent);
+      return intent_cb(store, user_ctx, obj, intent);
     }
     return 0;
   }
@@ -177,13 +183,66 @@ struct RGWPoolIterCtx {
   librados::IoCtx io_ctx;
   librados::ObjectIterator iter;
 };
+
+struct RGWRadosParams {
+  rgw_bucket domain_root;
+  rgw_bucket control_pool;
+  rgw_bucket gc_pool;
+  rgw_bucket log_pool;
+  rgw_bucket intent_log_pool;
+  rgw_bucket usage_log_pool;
+
+  rgw_bucket user_keys_pool;
+  rgw_bucket user_email_pool;
+  rgw_bucket user_swift_pool;
+  rgw_bucket user_uid_pool;
+
+  int init(CephContext *cct, RGWRados *store);
+  void init_default();
+
+  void encode(bufferlist& bl) const {
+    ENCODE_START(1, 1, bl);
+    ::encode(domain_root, bl);
+    ::encode(control_pool, bl);
+    ::encode(gc_pool, bl);
+    ::encode(log_pool, bl);
+    ::encode(intent_log_pool, bl);
+    ::encode(usage_log_pool, bl);
+    ::encode(user_keys_pool, bl);
+    ::encode(user_email_pool, bl);
+    ::encode(user_swift_pool, bl);
+    ::encode(user_uid_pool, bl);
+    ENCODE_FINISH(bl);
+  }
+
+  void decode(bufferlist::iterator& bl) {
+     DECODE_START(1, bl);
+    ::decode(domain_root, bl);
+    ::decode(control_pool, bl);
+    ::decode(gc_pool, bl);
+    ::decode(log_pool, bl);
+    ::decode(intent_log_pool, bl);
+    ::decode(usage_log_pool, bl);
+    ::decode(user_keys_pool, bl);
+    ::decode(user_email_pool, bl);
+    ::decode(user_swift_pool, bl);
+    ::decode(user_uid_pool, bl);
+    DECODE_FINISH(bl);
+  }
+  void dump(Formatter *f) const;
+};
+WRITE_CLASS_ENCODER(RGWRadosParams);
   
 class RGWRados
 {
+  friend class RGWGC;
+
   /** Open the pool used as root for this gateway */
   int open_root_pool_ctx();
+  int open_gc_pool_ctx();
 
   int open_bucket_ctx(rgw_bucket& bucket, librados::IoCtx&  io_ctx);
+  int open_bucket(rgw_bucket& bucket, librados::IoCtx&  io_ctx, string& bucket_oid);
 
   struct GetObjState {
     librados::IoCtx io_ctx;
@@ -192,8 +251,6 @@ class RGWRados
     GetObjState() : sent_data(false) {}
   };
 
-  int set_buckets_auid(vector<rgw_bucket>& buckets, uint64_t auid);
-
   Mutex lock;
   SafeTimer *timer;
 
@@ -206,22 +263,27 @@ class RGWRados
     }
   };
 
+  RGWGC *gc;
+  bool use_gc_thread;
 
-  RGWWatcher *watcher;
-  uint64_t watch_handle;
+  int num_watchers;
+  RGWWatcher **watchers;
+  uint64_t *watch_handles;
   librados::IoCtx root_pool_ctx;      // .rgw
   librados::IoCtx control_pool_ctx;   // .rgw.control
 
   Mutex bucket_id_lock;
   uint64_t max_bucket_id;
 
-  int get_obj_state(RGWRadosCtx *rctx, rgw_obj& obj, librados::IoCtx& io_ctx, string& actual_obj, RGWObjState **state);
-  int append_atomic_test(RGWRadosCtx *rctx, rgw_obj& obj, librados::IoCtx& io_ctx,
-                         string& actual_obj, librados::ObjectOperation& op, RGWObjState **state);
-  int prepare_atomic_for_write_impl(RGWRadosCtx *rctx, rgw_obj& obj, librados::IoCtx& io_ctx,
-                         string& actual_obj, librados::ObjectWriteOperation& op, RGWObjState **pstate);
-  int prepare_atomic_for_write(RGWRadosCtx *rctx, rgw_obj& obj, librados::IoCtx& io_ctx,
-                         string& actual_obj, librados::ObjectWriteOperation& op, RGWObjState **pstate);
+  int get_obj_state(RGWRadosCtx *rctx, rgw_obj& obj, RGWObjState **state);
+  int append_atomic_test(RGWRadosCtx *rctx, rgw_obj& obj,
+                         librados::ObjectOperation& op, RGWObjState **state);
+  int prepare_atomic_for_write_impl(RGWRadosCtx *rctx, rgw_obj& obj,
+                         librados::ObjectWriteOperation& op, RGWObjState **pstate,
+			 bool reset_obj, const string *ptag);
+  int prepare_atomic_for_write(RGWRadosCtx *rctx, rgw_obj& obj,
+                         librados::ObjectWriteOperation& op, RGWObjState **pstate,
+			 bool reset_obj, const string *ptag);
 
   void atomic_write_finish(RGWObjState *state, int r) {
     if (state && r == -ECANCELED) {
@@ -252,7 +314,7 @@ class RGWRados
     v.push_back(info);
     return clone_objs(ctx, dst_obj, v, attrs, category, pmtime, true, false);
   }
-  int delete_obj_impl(void *ctx, rgw_obj& src_obj, bool sync);
+  int delete_obj_impl(void *ctx, rgw_obj& src_obj);
   int complete_atomic_overwrite(RGWRadosCtx *rctx, RGWObjState *state, rgw_obj& obj);
 
   int update_placement_map();
@@ -261,27 +323,34 @@ class RGWRados
 
 protected:
   CephContext *cct;
+  librados::Rados *rados;
+  librados::IoCtx gc_pool_ctx;        // .rgw.gc
+
+  bool pools_initialized;
 
 public:
-  RGWRados() : lock("rados_timer_lock"), timer(NULL), watcher(NULL), watch_handle(0),
-               bucket_id_lock("rados_bucket_id"), max_bucket_id(0) {}
+  RGWRados() : lock("rados_timer_lock"), timer(NULL), gc(NULL), use_gc_thread(false),
+               num_watchers(0), watchers(NULL), watch_handles(NULL),
+               bucket_id_lock("rados_bucket_id"), max_bucket_id(0),
+               cct(NULL), rados(NULL),
+               pools_initialized(false) {}
+
+  RGWRadosParams params;
+
   virtual ~RGWRados() {}
 
   void tick();
 
   CephContext *ctx() { return cct; }
   /** do all necessary setup of the storage device */
-  int initialize(CephContext *_cct) {
+  int initialize(CephContext *_cct, bool _use_gc_thread) {
     cct = _cct;
+    use_gc_thread = _use_gc_thread;
     return initialize();
   }
   /** Initialize the RADOS instance and prepare to do other ops */
   virtual int initialize();
-  virtual void finalize() {}
-
-  static RGWRados *init_storage_provider(CephContext *cct);
-  static void close_storage();
-  static RGWRados *store;
+  virtual void finalize();
 
   /** set up a bucket listing. handle is filled in. */
   virtual int list_buckets_init(RGWAccessHandle *handle);
@@ -302,6 +371,11 @@ public:
   int log_show_init(const string& name, RGWAccessHandle *handle);
   int log_show_next(RGWAccessHandle handle, rgw_log_entry *entry);
 
+  // log bandwidth info
+  int log_usage(map<rgw_user_bucket, RGWUsageBatch>& usage_info);
+  int read_usage(string& user, uint64_t start_epoch, uint64_t end_epoch, uint32_t max_entries,
+                 bool *is_truncated, RGWUsageIter& read_iter, map<rgw_user_bucket, rgw_usage_log_entry>& usage);
+  int trim_usage(string& user, uint64_t start_epoch, uint64_t end_epoch);
 
   /**
    * get listing of the objects in a bucket.
@@ -320,24 +394,25 @@ public:
                    std::string& marker, std::vector<RGWObjEnt>& result, map<string, bool>& common_prefixes,
 		   bool get_content_type, string& ns, bool *is_truncated, RGWAccessListFilter *filter);
 
+  virtual int create_pool(rgw_bucket& bucket);
+
   /**
    * create a bucket with name bucket and the given list of attrs
    * returns 0 on success, -ERR# otherwise.
    */
   virtual int create_bucket(string& owner, rgw_bucket& bucket,
                             map<std::string,bufferlist>& attrs,
-                            bool system_bucket, bool exclusive = true,
-                            uint64_t auid = 0);
+                            bool exclusive = true);
   virtual int add_bucket_placement(std::string& new_pool);
   virtual int remove_bucket_placement(std::string& new_pool);
   virtual int list_placement_set(set<string>& names);
-  virtual int create_pools(vector<string>& names, vector<int>& retcodes, int auid = 0);
+  virtual int create_pools(vector<string>& names, vector<int>& retcodes);
 
   /** Write/overwrite an object to the bucket storage. */
   virtual int put_obj_meta(void *ctx, rgw_obj& obj, uint64_t size, time_t *mtime,
               map<std::string, bufferlist>& attrs, RGWObjCategory category, bool exclusive,
               map<std::string, bufferlist>* rmattrs, const bufferlist *data,
-              RGWObjManifest *manifest);
+              RGWObjManifest *manifest, const string *ptag);
   virtual int put_obj_data(void *ctx, rgw_obj& obj, const char *data,
               off_t ofs, size_t len, bool exclusive);
   virtual int aio_put_obj_data(void *ctx, rgw_obj& obj, bufferlist& bl,
@@ -347,7 +422,7 @@ public:
               time_t *mtime, map<std::string, bufferlist>& attrs) {
     bufferlist bl;
     bl.append(data, len);
-    int ret = put_obj_meta(ctx, obj, len, mtime, attrs, RGW_OBJ_CATEGORY_NONE, exclusive, NULL, &bl, NULL);
+    int ret = put_obj_meta(ctx, obj, len, mtime, attrs, RGW_OBJ_CATEGORY_NONE, exclusive, NULL, &bl, NULL, NULL);
     return ret;
   }
   virtual int aio_wait(void *handle);
@@ -389,13 +464,9 @@ public:
 
   /**
    * Copy an object.
-   * dest_bucket: the bucket to copy into
    * dest_obj: the object to copy into
-   * src_bucket: the bucket to copy from
    * src_obj: the object to copy from
-   * mod_ptr, unmod_ptr, if_match, if_nomatch: as used in get_obj
-   * attrs: these are placed on the new object IN ADDITION to
-   *    (or overwriting) any attrs copied from the original object
+   * attrs: if replace_attrs is set then these are placed on the new object
    * err: stores any errors resulting from the get of the original object
    * Returns: 0 on success, -ERR# otherwise.
    */
@@ -406,9 +477,19 @@ public:
                const time_t *unmod_ptr,
                const char *if_match,
                const char *if_nomatch,
+               bool replace_attrs,
                map<std::string, bufferlist>& attrs,
                RGWObjCategory category,
                struct rgw_err *err);
+
+  int copy_obj_data(void *ctx,
+	       void *handle, off_t end,
+               rgw_obj& dest_obj,
+               rgw_obj& src_obj,
+	       time_t *mtime,
+               map<string, bufferlist>& attrs,
+               RGWObjCategory category,
+               struct rgw_err *err);
   /**
    * Delete a bucket.
    * bucket: the name of the bucket to delete
@@ -419,7 +500,7 @@ public:
   virtual int bucket_suspended(rgw_bucket& bucket, bool *suspended);
 
   /** Delete an object.*/
-  virtual int delete_obj(void *ctx, rgw_obj& src_obj, bool sync = true);
+  virtual int delete_obj(void *ctx, rgw_obj& src_obj);
 
   /**
    * Get the attributes for an object.
@@ -490,7 +571,8 @@ public:
    */
   virtual int read(void *ctx, rgw_obj& obj, off_t ofs, size_t size, bufferlist& bl);
 
-  virtual int obj_stat(void *ctx, rgw_obj& obj, uint64_t *psize, time_t *pmtime, map<string, bufferlist> *attrs, bufferlist *first_chunk);
+  virtual int obj_stat(void *ctx, rgw_obj& obj, uint64_t *psize, time_t *pmtime,
+                       uint64_t *epoch, map<string, bufferlist> *attrs, bufferlist *first_chunk);
 
   virtual bool supports_omap() { return true; }
   virtual int omap_get_all(rgw_obj& obj, bufferlist& header, std::map<string, bufferlist>& m);
@@ -502,11 +584,12 @@ public:
 
   virtual int init_watch();
   virtual void finalize_watch();
-  virtual int distribute(bufferlist& bl);
+  virtual int distribute(const string& key, bufferlist& bl);
   virtual int watch_cb(int opcode, uint64_t ver, bufferlist& bl) { return 0; }
+  void pick_control_oid(const string& key, string& notify_oid);
 
   void *create_context(void *user_ctx) {
-    RGWRadosCtx *rctx = new RGWRadosCtx();
+    RGWRadosCtx *rctx = new RGWRadosCtx(this);
     rctx->user_ctx = user_ctx;
     return rctx;
   }
@@ -523,15 +606,15 @@ public:
   }
   // to notify upper layer that we need to do some operation on an object, and it's up to
   // the upper layer to schedule this operation.. e.g., log intent in intent log
-  void set_intent_cb(void *ctx, int (*cb)(void *user_ctx, rgw_obj& obj, RGWIntentEvent intent)) {
+  void set_intent_cb(void *ctx, int (*cb)(RGWRados *store, void *user_ctx, rgw_obj& obj, RGWIntentEvent intent)) {
     RGWRadosCtx *rctx = (RGWRadosCtx *)ctx;
     rctx->set_intent_cb(cb);
   }
 
   int decode_policy(bufferlist& bl, ACLOwner *owner);
   int get_bucket_stats(rgw_bucket& bucket, map<RGWObjCategory, RGWBucketStats>& stats);
-  virtual int get_bucket_info(void *ctx, string& bucket_name, RGWBucketInfo& info);
-  virtual int put_bucket_info(string& bucket_name, RGWBucketInfo& info, bool exclusive);
+  virtual int get_bucket_info(void *ctx, string& bucket_name, RGWBucketInfo& info, map<string, bufferlist> *pattrs = NULL);
+  virtual int put_bucket_info(string& bucket_name, RGWBucketInfo& info, bool exclusive, map<string, bufferlist> *pattrs);
 
   int cls_rgw_init_index(librados::IoCtx& io_ctx, librados::ObjectWriteOperation& op, string& oid);
   int cls_obj_prepare_op(rgw_bucket& bucket, uint8_t op, string& tag,
@@ -556,9 +639,26 @@ public:
     return cls_obj_complete_cancel(bucket, tag, oid);
   }
 
+  int cls_obj_usage_log_add(const string& oid, rgw_usage_log_info& info);
+  int cls_obj_usage_log_read(string& oid, string& user, uint64_t start_epoch, uint64_t end_epoch, uint32_t max_entries,
+                             string& read_iter, map<rgw_user_bucket, rgw_usage_log_entry>& usage, bool *is_truncated);
+  int cls_obj_usage_log_trim(string& oid, string& user, uint64_t start_epoch, uint64_t end_epoch);
+
   /// clean up/process any temporary objects older than given date[/time]
   int remove_temp_objects(string date, string time);
 
+  int gc_operate(string& oid, librados::ObjectWriteOperation *op);
+  int gc_aio_operate(string& oid, librados::ObjectWriteOperation *op);
+  int gc_operate(string& oid, librados::ObjectReadOperation *op, bufferlist *pbl);
+
+  int list_gc_objs(int *index, string& marker, uint32_t max, std::list<cls_rgw_gc_obj_info>& result, bool *truncated);
+  int process_gc();
+  int defer_gc(void *ctx, rgw_obj& obj);
+
+  int bucket_check_index(rgw_bucket& bucket,
+                         map<RGWObjCategory, RGWBucketStats> *existing_stats,
+                         map<RGWObjCategory, RGWBucketStats> *calculated_stats);
+  int bucket_rebuild_index(rgw_bucket& bucket);
  private:
   int process_intent_log(rgw_bucket& bucket, string& oid,
 			 time_t epoch, int flags, bool purge);
@@ -607,23 +707,21 @@ public:
 
   uint64_t instance_id();
   uint64_t next_bucket_id();
+
 };
 
 class RGWStoreManager {
-  RGWRados *store;
 public:
-  RGWStoreManager() : store(NULL) {}
-  ~RGWStoreManager() {
-    if (store)
-      RGWRados::close_storage();
-  }
-  RGWRados *init(CephContext *cct) {
-    store = RGWRados::init_storage_provider(cct);
+  RGWStoreManager() {}
+  static RGWRados *get_storage(CephContext *cct, bool use_gc_thread) {
+    RGWRados *store = init_storage_provider(cct, use_gc_thread);
     return store;
   }
+  static RGWRados *init_storage_provider(CephContext *cct, bool use_gc_thread);
+  static void close_storage(RGWRados *store);
+
 };
 
-#define rgwstore RGWRados::store
 
 
 #endif
