@@ -34,13 +34,28 @@ void list_all_buckets_end(struct req_state *s)
 void dump_bucket(struct req_state *s, RGWBucketEnt& obj)
 {
   s->formatter->open_object_section("Bucket");
-  s->formatter->dump_format("Name", obj.bucket.name.c_str());
+  s->formatter->dump_string("Name", obj.bucket.name);
   dump_time(s, "CreationDate", &obj.mtime);
   s->formatter->close_section();
 }
 
+void rgw_get_errno_s3(rgw_html_errors *e , int err_no)
+{
+  const struct rgw_html_errors *r;
+  r = search_err(err_no, RGW_HTML_ERRORS, ARRAY_LEN(RGW_HTML_ERRORS));
+
+  if (r) {
+    e->http_ret = r->http_ret;
+    e->s3_code = r->s3_code;
+  } else {
+    e->http_ret = 500;
+    e->s3_code = "UnknownError";
+  }
+}
+
 int RGWGetObj_REST_S3::send_response(bufferlist& bl)
 {
+  string content_type_str;
   const char *content_type = NULL;
   int orig_ret = ret;
 
@@ -66,6 +81,28 @@ int RGWGetObj_REST_S3::send_response(bufferlist& bl)
       }
     }
 
+    if (s->args.has_response_modifier()) {
+      bool exists;
+      content_type_str = s->args.get("response-content-type", &exists);
+      if (exists)
+	content_type = content_type_str.c_str();
+      string val = s->args.get("response-content-language", &exists);
+      if (exists)
+        CGI_PRINTF(s, "Content-Language: %s\n", val.c_str());
+      val = s->args.get("response-expires", &exists);
+      if (exists)
+        CGI_PRINTF(s, "Expires: %s\n", val.c_str());
+      val = s->args.get("response-cache-control", &exists);
+      if (exists)
+        CGI_PRINTF(s, "Cache-Control: %s\n", val.c_str());
+      val = s->args.get("response-content-disposition", &exists);
+      if (exists)
+        CGI_PRINTF(s, "Content-Disposition: %s\n", val.c_str());
+      val = s->args.get("response-content-encoding", &exists);
+      if (exists)
+        CGI_PRINTF(s, "Content-Encoding: %s\n", val.c_str());
+    }
+
     for (iter = attrs.begin(); iter != attrs.end(); ++iter) {
        const char *name = iter->first.c_str();
        if (strncmp(name, RGW_ATTR_META_PREFIX, sizeof(RGW_ATTR_META_PREFIX)-1) == 0) {
@@ -118,7 +155,7 @@ void RGWListBuckets_REST_S3::send_response()
   list_all_buckets_end(s);
   dump_content_length(s, s->formatter->get_len());
   end_header(s, "application/xml");
-  flush_formatter_to_req_state(s, s->formatter);
+  rgw_flush_formatter_and_reset(s, s->formatter);
 }
 
 int RGWListBucket_REST_S3::get_params()
@@ -147,25 +184,25 @@ void RGWListBucket_REST_S3::send_response()
 
   s->formatter->open_object_section_in_ns("ListBucketResult",
 					  "http://s3.amazonaws.com/doc/2006-03-01/");
-  s->formatter->dump_format("Name", s->bucket_name);
+  s->formatter->dump_string("Name", s->bucket_name);
   if (!prefix.empty())
-    s->formatter->dump_format("Prefix", prefix.c_str());
-  s->formatter->dump_format("Marker", marker.c_str());
+    s->formatter->dump_string("Prefix", prefix);
+  s->formatter->dump_string("Marker", marker);
   s->formatter->dump_int("MaxKeys", max);
   if (!delimiter.empty())
-    s->formatter->dump_format("Delimiter", delimiter.c_str());
+    s->formatter->dump_string("Delimiter", delimiter);
 
-  s->formatter->dump_format("IsTruncated", (max && is_truncated ? "true" : "false"));
+  s->formatter->dump_string("IsTruncated", (max && is_truncated ? "true" : "false"));
 
   if (ret >= 0) {
     vector<RGWObjEnt>::iterator iter;
     for (iter = objs.begin(); iter != objs.end(); ++iter) {
       s->formatter->open_array_section("Contents");
-      s->formatter->dump_string("Key", iter->name.c_str());
+      s->formatter->dump_string("Key", iter->name);
       dump_time(s, "LastModified", &iter->mtime);
       s->formatter->dump_format("ETag", "\"%s\"", iter->etag.c_str());
       s->formatter->dump_int("Size", iter->size);
-      s->formatter->dump_format("StorageClass", "STANDARD");
+      s->formatter->dump_string("StorageClass", "STANDARD");
       dump_owner(s, iter->owner, iter->owner_display_name);
       s->formatter->close_section();
     }
@@ -173,13 +210,13 @@ void RGWListBucket_REST_S3::send_response()
       map<string, bool>::iterator pref_iter;
       for (pref_iter = common_prefixes.begin(); pref_iter != common_prefixes.end(); ++pref_iter) {
         s->formatter->open_array_section("CommonPrefixes");
-        s->formatter->dump_format("Prefix", pref_iter->first.c_str());
+        s->formatter->dump_string("Prefix", pref_iter->first);
         s->formatter->close_section();
       }
     }
   }
   s->formatter->close_section();
-  flush_formatter_to_req_state(s, s->formatter);
+  rgw_flush_formatter_and_reset(s, s->formatter);
 }
 
 static void dump_bucket_metadata(struct req_state *s, RGWBucketEnt& bucket)
@@ -309,6 +346,23 @@ int RGWCopyObj_REST_S3::get_params()
   dest_bucket_name = s->bucket.name;
   dest_object = s->object_str;
 
+  const char *md_directive = s->env->get("HTTP_X_AMZ_METADATA_DIRECTIVE");
+  if (md_directive) {
+    if (strcasecmp(md_directive, "COPY") == 0) {
+      replace_attrs = false;
+    } else if (strcasecmp(md_directive, "REPLACE") == 0) {
+      replace_attrs = true;
+    } else {
+      return -EINVAL;
+    }
+  }
+
+  if ((dest_bucket_name.compare(src_bucket_name) == 0) &&
+      (dest_object.compare(src_object) == 0) &&
+      !replace_attrs) {
+    /* can only copy object into itself if replacing attrs */
+    return -ERR_INVALID_REQUEST;
+  }
   return 0;
 }
 
@@ -327,11 +381,11 @@ void RGWCopyObj_REST_S3::send_response()
       bufferlist& bl = iter->second;
       if (bl.length()) {
         char *etag = bl.c_str();
-        s->formatter->dump_format("ETag", etag);
+        s->formatter->dump_string("ETag", etag);
       }
     }
     s->formatter->close_section();
-    flush_formatter_to_req_state(s, s->formatter);
+    rgw_flush_formatter_and_reset(s, s->formatter);
   }
 }
 
@@ -388,11 +442,11 @@ void RGWInitMultipart_REST_S3::send_response()
     dump_start(s);
     s->formatter->open_object_section_in_ns("InitiateMultipartUploadResult",
 		  "http://s3.amazonaws.com/doc/2006-03-01/");
-    s->formatter->dump_format("Bucket", s->bucket_name);
-    s->formatter->dump_format("Key", s->object);
-    s->formatter->dump_format("UploadId", upload_id.c_str());
+    s->formatter->dump_string("Bucket", s->bucket_name);
+    s->formatter->dump_string("Key", s->object);
+    s->formatter->dump_string("UploadId", upload_id);
     s->formatter->close_section();
-    flush_formatter_to_req_state(s, s->formatter);
+    rgw_flush_formatter_and_reset(s, s->formatter);
   }
 }
 
@@ -408,11 +462,11 @@ void RGWCompleteMultipart_REST_S3::send_response()
 			  "http://s3.amazonaws.com/doc/2006-03-01/");
     if (g_conf->rgw_dns_name.length())
       s->formatter->dump_format("Location", "%s.%s", s->bucket_name, g_conf->rgw_dns_name.c_str());
-    s->formatter->dump_format("Bucket", s->bucket_name);
-    s->formatter->dump_format("Key", s->object);
-    s->formatter->dump_format("ETag", etag.c_str());
+    s->formatter->dump_string("Bucket", s->bucket_name);
+    s->formatter->dump_string("Key", s->object);
+    s->formatter->dump_string("ETag", etag);
     s->formatter->close_section();
-    flush_formatter_to_req_state(s, s->formatter);
+    rgw_flush_formatter_and_reset(s, s->formatter);
   }
 }
 
@@ -445,14 +499,14 @@ void RGWListMultipart_REST_S3::send_response()
     for (i = 0, test_iter = iter; test_iter != parts.end() && i < max_parts; ++test_iter, ++i) {
       cur_max = test_iter->first;
     }
-    s->formatter->dump_format("Bucket", s->bucket_name);
-    s->formatter->dump_format("Key", s->object);
-    s->formatter->dump_format("UploadId", upload_id.c_str());
-    s->formatter->dump_format("StorageClass", "STANDARD");
-    s->formatter->dump_format("PartNumberMarker", "%d", marker);
-    s->formatter->dump_format("NextPartNumberMarker", "%d", cur_max + 1);
-    s->formatter->dump_format("MaxParts", "%d", max_parts);
-    s->formatter->dump_format("IsTruncated", "%s", (test_iter == parts.end() ? "false" : "true"));
+    s->formatter->dump_string("Bucket", s->bucket_name);
+    s->formatter->dump_string("Key", s->object);
+    s->formatter->dump_string("UploadId", upload_id);
+    s->formatter->dump_string("StorageClass", "STANDARD");
+    s->formatter->dump_int("PartNumberMarker", marker);
+    s->formatter->dump_int("NextPartNumberMarker", cur_max + 1);
+    s->formatter->dump_int("MaxParts", max_parts);
+    s->formatter->dump_string("IsTruncated", (test_iter == parts.end() ? "false" : "true"));
 
     ACLOwner& owner = policy.get_owner();
     dump_owner(s, owner.get_id(), owner.get_display_name());
@@ -468,16 +522,16 @@ void RGWListMultipart_REST_S3::send_response()
       s->formatter->open_object_section("Part");
 
       if (strftime(buf, sizeof(buf), "%Y-%m-%dT%T.000Z", &tmp) > 0) {
-        s->formatter->dump_format("LastModified", buf);
+        s->formatter->dump_string("LastModified", buf);
       }
 
       s->formatter->dump_unsigned("PartNumber", info.num);
-      s->formatter->dump_format("ETag", "%s", info.etag.c_str());
+      s->formatter->dump_string("ETag", info.etag);
       s->formatter->dump_unsigned("Size", info.size);
       s->formatter->close_section();
     }
     s->formatter->close_section();
-    flush_formatter_to_req_state(s, s->formatter);
+    rgw_flush_formatter_and_reset(s, s->formatter);
   }
 }
 
@@ -493,36 +547,36 @@ void RGWListBucketMultiparts_REST_S3::send_response()
     return;
 
   s->formatter->open_object_section("ListMultipartUploadsResult");
-  s->formatter->dump_format("Bucket", s->bucket_name);
+  s->formatter->dump_string("Bucket", s->bucket_name);
   if (!prefix.empty())
-    s->formatter->dump_format("ListMultipartUploadsResult.Prefix", prefix.c_str());
+    s->formatter->dump_string("ListMultipartUploadsResult.Prefix", prefix);
   string& key_marker = marker.get_key();
   if (!key_marker.empty())
-    s->formatter->dump_format("KeyMarker", key_marker.c_str());
+    s->formatter->dump_string("KeyMarker", key_marker);
   string& upload_id_marker = marker.get_upload_id();
   if (!upload_id_marker.empty())
-    s->formatter->dump_format("UploadIdMarker", upload_id_marker.c_str());
+    s->formatter->dump_string("UploadIdMarker", upload_id_marker);
   string next_key = next_marker.mp.get_key();
   if (!next_key.empty())
-    s->formatter->dump_format("NextKeyMarker", next_key.c_str());
+    s->formatter->dump_string("NextKeyMarker", next_key);
   string next_upload_id = next_marker.mp.get_upload_id();
   if (!next_upload_id.empty())
-    s->formatter->dump_format("NextUploadIdMarker", next_upload_id.c_str());
-  s->formatter->dump_format("MaxUploads", "%d", max_uploads);
+    s->formatter->dump_string("NextUploadIdMarker", next_upload_id);
+  s->formatter->dump_int("MaxUploads", max_uploads);
   if (!delimiter.empty())
-    s->formatter->dump_format("Delimiter", delimiter.c_str());
-  s->formatter->dump_format("IsTruncated", (is_truncated ? "true" : "false"));
+    s->formatter->dump_string("Delimiter", delimiter);
+  s->formatter->dump_string("IsTruncated", (is_truncated ? "true" : "false"));
 
   if (ret >= 0) {
     vector<RGWMultipartUploadEntry>::iterator iter;
     for (iter = uploads.begin(); iter != uploads.end(); ++iter) {
       RGWMPObj& mp = iter->mp;
       s->formatter->open_array_section("Upload");
-      s->formatter->dump_format("Key", mp.get_key().c_str());
-      s->formatter->dump_format("UploadId", mp.get_upload_id().c_str());
+      s->formatter->dump_string("Key", mp.get_key());
+      s->formatter->dump_string("UploadId", mp.get_upload_id());
       dump_owner(s, s->user.user_id, s->user.display_name, "Initiator");
       dump_owner(s, s->user.user_id, s->user.display_name);
-      s->formatter->dump_format("StorageClass", "STANDARD");
+      s->formatter->dump_string("StorageClass", "STANDARD");
       dump_time(s, "Initiated", &iter->obj.mtime);
       s->formatter->close_section();
     }
@@ -530,13 +584,71 @@ void RGWListBucketMultiparts_REST_S3::send_response()
       s->formatter->open_array_section("CommonPrefixes");
       map<string, bool>::iterator pref_iter;
       for (pref_iter = common_prefixes.begin(); pref_iter != common_prefixes.end(); ++pref_iter) {
-        s->formatter->dump_format("CommonPrefixes.Prefix", pref_iter->first.c_str());
+        s->formatter->dump_string("CommonPrefixes.Prefix", pref_iter->first);
       }
       s->formatter->close_section();
     }
   }
   s->formatter->close_section();
-  flush_formatter_to_req_state(s, s->formatter);
+  rgw_flush_formatter_and_reset(s, s->formatter);
+}
+
+void RGWDeleteMultiObj_REST_S3::send_status()
+{
+  if (!status_dumped) {
+    if (ret < 0)
+      set_req_state_err(s, ret);
+    dump_errno(s);
+    status_dumped = true;
+  }
+}
+
+void RGWDeleteMultiObj_REST_S3::begin_response()
+{
+
+  if (!status_dumped) {
+    send_status();
+  }
+
+  dump_start(s);
+  end_header(s, "application/xml");
+  s->formatter->open_object_section_in_ns("DeleteResult",
+                                            "http://s3.amazonaws.com/doc/2006-03-01/");
+
+  rgw_flush_formatter(s, s->formatter);
+}
+
+void RGWDeleteMultiObj_REST_S3::send_partial_response(pair<string,int>& result)
+{
+  if (!result.first.empty()) {
+    if (result.second == 0 && !quiet) {
+      s->formatter->open_object_section("Deleted");
+      s->formatter->dump_string("Key", result.first);
+      s->formatter->close_section();
+    } else if (result.first < 0) {
+      struct rgw_html_errors *r = new rgw_html_errors;
+      int err_no;
+
+      s->formatter->open_object_section("Error");
+
+      err_no = -(result.second);
+      rgw_get_errno_s3(r, err_no);
+
+      s->formatter->dump_string("Key", result.first);
+      s->formatter->dump_int("Code", r->http_ret);
+      s->formatter->dump_string("Message", r->s3_code);
+      s->formatter->close_section();
+    }
+
+    rgw_flush_formatter(s, s->formatter);
+  }
+}
+
+void RGWDeleteMultiObj_REST_S3::end_response()
+{
+
+  s->formatter->close_section();
+  rgw_flush_formatter_and_reset(s, s->formatter);
 }
 
 RGWOp *RGWHandler_REST_S3::get_retrieve_obj_op(bool get_data)
@@ -614,11 +726,14 @@ RGWOp *RGWHandler_REST_S3::get_post_op()
     else
       return new RGWInitMultipart_REST_S3;
   }
+  else if (s->args.sub_resource_exists("delete")) {
+    return new RGWDeleteMultiObj_REST_S3;
+  }
 
   return NULL;
 }
 
-int RGWHandler_REST_S3::init(struct req_state *state, FCGX_Request *fcgx)
+int RGWHandler_REST_S3::init(RGWRados *store, struct req_state *state, FCGX_Request *fcgx)
 {
   const char *cacl = state->env->get("HTTP_X_AMZ_ACL");
   if (cacl)
@@ -628,7 +743,7 @@ int RGWHandler_REST_S3::init(struct req_state *state, FCGX_Request *fcgx)
 
   state->dialect = "s3";
 
-  return RGWHandler_REST::init(state, fcgx);
+  return RGWHandler_REST::init(store, state, fcgx);
 }
 
 /*
@@ -793,7 +908,7 @@ int RGWHandler_REST_S3::authorize()
   }
 
   /* first get the user info */
-  if (rgw_get_user_info_by_access_key(auth_id, s->user) < 0) {
+  if (rgw_get_user_info_by_access_key(store, auth_id, s->user) < 0) {
     dout(5) << "error reading user info, uid=" << auth_id << " can't authenticate" << dendl;
     return -EPERM;
   }
