@@ -67,7 +67,7 @@ done:
     return;
   }
 
-  flush_formatter_to_req_state(s, s->formatter);
+  rgw_flush_formatter_and_reset(s, s->formatter);
 }
 
 int RGWListBucket_REST_SWIFT::get_params()
@@ -186,7 +186,7 @@ next:
     return;
   }
 
-  flush_formatter_to_req_state(s, s->formatter);
+  rgw_flush_formatter_and_reset(s, s->formatter);
 }
 
 static void dump_container_metadata(struct req_state *s, RGWBucketEnt& bucket)
@@ -270,7 +270,7 @@ void RGWCreateBucket_REST_SWIFT::send_response()
   set_req_state_err(s, ret);
   dump_errno(s);
   end_header(s);
-  flush_formatter_to_req_state(s, s->formatter);
+  rgw_flush_formatter_and_reset(s, s->formatter);
 }
 
 void RGWDeleteBucket_REST_SWIFT::send_response()
@@ -282,7 +282,7 @@ void RGWDeleteBucket_REST_SWIFT::send_response()
   set_req_state_err(s, r);
   dump_errno(s);
   end_header(s);
-  flush_formatter_to_req_state(s, s->formatter);
+  rgw_flush_formatter_and_reset(s, s->formatter);
 }
 
 int RGWPutObj_REST_SWIFT::get_params()
@@ -314,6 +314,8 @@ int RGWPutObj_REST_SWIFT::get_params()
 
   policy.create_default(s->user.user_id, s->user.display_name);
 
+  obj_manifest = s->env->get("HTTP_X_OBJECT_MANIFEST");
+
   return RGWPutObj_REST::get_params();
 }
 
@@ -325,7 +327,7 @@ void RGWPutObj_REST_SWIFT::send_response()
   set_req_state_err(s, ret);
   dump_errno(s);
   end_header(s);
-  flush_formatter_to_req_state(s, s->formatter);
+  rgw_flush_formatter_and_reset(s, s->formatter);
 }
 
 int RGWPutMetadata_REST_SWIFT::get_params()
@@ -347,7 +349,7 @@ int RGWPutMetadata_REST_SWIFT::get_params()
 
     if (read_attr || write_attr) {
       RGWAccessControlPolicy_SWIFT swift_policy(s->cct);
-      int r = swift_policy.create(s->user.user_id, s->user.display_name, read_list, write_list);
+      int r = swift_policy.create(store, s->user.user_id, s->user.display_name, read_list, write_list);
       if (r < 0)
         return r;
 
@@ -366,7 +368,7 @@ void RGWPutMetadata_REST_SWIFT::send_response()
   set_req_state_err(s, ret);
   dump_errno(s);
   end_header(s);
-  flush_formatter_to_req_state(s, s->formatter);
+  rgw_flush_formatter_and_reset(s, s->formatter);
 }
 
 void RGWDeleteObj_REST_SWIFT::send_response()
@@ -378,7 +380,7 @@ void RGWDeleteObj_REST_SWIFT::send_response()
   set_req_state_err(s, r);
   dump_errno(s);
   end_header(s);
-  flush_formatter_to_req_state(s, s->formatter);
+  rgw_flush_formatter_and_reset(s, s->formatter);
 }
 
 int RGWCopyObj_REST_SWIFT::init_dest_policy()
@@ -481,7 +483,7 @@ send_data:
   if (get_data && !orig_ret) {
     CGI_PutStr(s, bl.c_str(), len);
   }
-  flush_formatter_to_req_state(s, s->formatter);
+  rgw_flush_formatter_and_reset(s, s->formatter);
 
   return 0;
 }
@@ -562,7 +564,7 @@ RGWOp *RGWHandler_REST_SWIFT::get_copy_op()
 
 int RGWHandler_REST_SWIFT::authorize()
 {
-  bool authorized = rgw_verify_os_token(s);
+  bool authorized = rgw_verify_swift_token(store, s);
   if (!authorized)
     return -EPERM;
 
@@ -571,11 +573,11 @@ int RGWHandler_REST_SWIFT::authorize()
   return 0;
 }
 
-int RGWHandler_REST_SWIFT::init(struct req_state *state, FCGX_Request *fcgx)
+int RGWHandler_REST_SWIFT::init(RGWRados *store, struct req_state *state, FCGX_Request *fcgx)
 {
   state->copy_source = state->env->get("HTTP_X_COPY_FROM");
 
   state->dialect = "swift";
 
-  return RGWHandler_REST::init(state, fcgx);
+  return RGWHandler_REST::init(store, state, fcgx);
 }
