@@ -17,21 +17,6 @@
 #include <string>
 using namespace std;
 
-#include "acconfig.h"
-#include "messages/MMonCommand.h"
-#include "messages/MMonCommandAck.h"
-#include "mon/MonClient.h"
-#include "mon/MonMap.h"
-#include "osd/OSDMap.h"
-#include "msg/SimpleMessenger.h"
-#include "tools/common.h"
-
-#include "common/errno.h"
-#include "common/Cond.h"
-#include "common/Mutex.h"
-#include "common/Timer.h"
-#include "global/global_init.h"
-
 #if !defined(DARWIN) && !defined(__FreeBSD__)
 #include <envz.h>
 #endif // DARWIN
@@ -45,6 +30,25 @@ extern "C" {
 #include <histedit.h>
 }
 
+#include "acconfig.h"
+#include "messages/MMonCommand.h"
+#include "messages/MMonCommandAck.h"
+#include "messages/MLog.h"
+#include "mon/MonClient.h"
+#include "mon/MonMap.h"
+#include "osd/OSDMap.h"
+#include "msg/SimpleMessenger.h"
+#include "tools/common.h"
+
+#include "common/errno.h"
+#include "common/Cond.h"
+#include "common/Mutex.h"
+#include "common/Timer.h"
+#include "global/global_init.h"
+
+
+#include "include/assert.h"
+
 #define dout_subsys ceph_subsys_
 
 // TODO: should move these into CephToolCtx for consistency
@@ -60,6 +64,7 @@ bool pending_tell_pgid;
 uint64_t pending_tid = 0;
 EntityName pending_target;
 pg_t pending_target_pgid;
+bool cmd_waiting_for_osdmap = false;
 vector<string> pending_cmd;
 bufferlist pending_bl;
 bool reply;
@@ -67,10 +72,13 @@ string reply_rs;
 int reply_rc;
 bufferlist reply_bl;
 entity_inst_t reply_from;
-Context *resend_event = 0;
 
 OSDMap *osdmap = 0;
 
+Connection *command_con = NULL;
+Context *tick_event = 0;
+float tick_interval = 3.0;
+
 // observe (push)
 #include "mon/PGMap.h"
 #include "osd/OSDMap.h"
@@ -79,8 +87,6 @@ OSDMap *osdmap = 0;
 
 #include "mon/mon_types.h"
 
-#include "messages/MMonObserve.h"
-#include "messages/MMonObserveNotify.h"
 #include "messages/MOSDMap.h"
 
 #include "messages/MCommand.h"
@@ -88,225 +94,19 @@ OSDMap *osdmap = 0;
 
 static set<int> registered, seen;
 
-version_t map_ver[PAXOS_NUM];
-
-static void handle_osd_map(CephToolCtx *ctx, MOSDMap *m)
-{
-  epoch_t e = m->get_first();
-  assert(m->maps.count(e));
-  ctx->lock.Lock();
-  delete osdmap;
-  osdmap = new OSDMap;
-  osdmap->decode(m->maps[e]);
-  cmd_cond.Signal();
-  ctx->lock.Unlock();
-  m->put();
-}
-
-static void handle_observe(CephToolCtx *ctx, MMonObserve *observe)
-{
-  dout(1) << observe->get_source() << " -> " << get_paxos_name(observe->machine_id)
-	  << " registered" << dendl;
-  ctx->lock.Lock();
-  registered.insert(observe->machine_id);  
-  ctx->lock.Unlock();
-  observe->put();
-}
-
-static void handle_notify(CephToolCtx *ctx, MMonObserveNotify *notify)
-{
-  utime_t now = ceph_clock_now(g_ceph_context);
-
-  dout(1) << notify->get_source() << " -> " << get_paxos_name(notify->machine_id)
-	  << " v" << notify->ver
-	  << (notify->is_latest ? " (latest)" : "")
-	  << dendl;
-  
-  if (notify->fsid != ctx->mc.monmap.fsid) {
-    dout(0) << notify->get_source_inst() << " notify fsid " << notify->fsid << " != "
-	    << ctx->mc.monmap.fsid << dendl;
-    notify->put();
-    return;
-  }
-
-  if (map_ver[notify->machine_id] >= notify->ver)
-    return;
-
-  switch (notify->machine_id) {
-  case PAXOS_PGMAP:
-    {
-      bufferlist::iterator p = notify->bl.begin();
-      if (notify->is_latest) {
-	ctx->pgmap.decode(p);
-      } else {
-	PGMap::Incremental inc;
-	inc.decode(p);
-	ctx->pgmap.apply_incremental(inc);
-      }
-      *ctx->log << now << "    pg " << ctx->pgmap << std::endl;
-      ctx->updates |= PG_MON_UPDATE;
-      break;
-    }
-
-  case PAXOS_MDSMAP:
-    ctx->mdsmap.decode(notify->bl);
-    *ctx->log << now << "   mds " << ctx->mdsmap << std::endl;
-    ctx->updates |= MDS_MON_UPDATE;
-    break;
-
-  case PAXOS_OSDMAP:
-    {
-      if (notify->is_latest) {
-	ctx->osdmap.decode(notify->bl);
-      } else {
-	OSDMap::Incremental inc(notify->bl);
-	ctx->osdmap.apply_incremental(inc);
-      }
-      *ctx->log << now << "   osd " << ctx->osdmap << std::endl;
-    }
-    ctx->updates |= OSD_MON_UPDATE;
-    break;
-
-  case PAXOS_LOG:
-    {
-      bufferlist::iterator p = notify->bl.begin();
-      if (notify->is_latest) {
-	LogSummary summary;
-	::decode(summary, p);
-	// show last log message
-	if (!summary.tail.empty())
-	  *ctx->log << now << "   log " << summary.tail.back() << std::endl;
-      } else {
-	LogEntry le;
-	__u8 v;
-	::decode(v, p);
-	while (!p.end()) {
-	  le.decode(p);
-	  *ctx->log << le << std::endl;
-	}
-      }
-      break;
-    }
-
-  case PAXOS_AUTH:
-    {
-#if 0
-      bufferlist::iterator p = notify->bl.begin();
-      if (notify->is_latest) {
-	KeyServerData ctx;
-	::decode(ctx, p);
-	*ctx->log << now << "   auth " << std::endl;
-      } else {
-	while (!p.end()) {
-	  AuthMonitor::Incremental inc;
-          inc.decode(p);
-	  *ctx->log << now << "   auth " << inc.name.to_str() << std::endl;
-	}
-      }
-#endif
-      /* ignoring auth incremental.. don't want to decode it */
-      break;
-    }
-
-  case PAXOS_MONMAP:
-    {
-      ctx->mc.monmap.decode(notify->bl);
-      *ctx->log << now << "   mon " << ctx->mc.monmap << std::endl;
-    }
-    break;
-
-  default:
-    *ctx->log << now << "  ignoring unknown machine id " << notify->machine_id << std::endl;
-  }
-
-  map_ver[notify->machine_id] = notify->ver;
-
-  // have we seen them all?
-  seen.insert(notify->machine_id);
-  switch (ceph_tool_mode) {
-    case CEPH_TOOL_MODE_ONE_SHOT_OBSERVER:
-      if (seen.size() == PAXOS_NUM) {
-	messenger->shutdown();
-      }
-      break;
-    case CEPH_TOOL_MODE_GUI:
-      ctx->gui_cond.Signal();
-      break;
-    default:
-      // do nothing
-      break;
-  }
-
-  notify->put();
-}
-
-class C_ObserverRefresh : public Context {
-public:
-  bool newmon;
-  C_ObserverRefresh(bool n, CephToolCtx *ctx_) 
-    : newmon(n),
-      ctx(ctx_)
-  {
-  }
-  void finish(int r) {
-    send_observe_requests(ctx);
-  }
-private:
+struct C_Tick : public Context {
   CephToolCtx *ctx;
+  C_Tick(CephToolCtx *c) : ctx(c) {}
+  void finish(int r) {
+    if (command_con)
+      messenger->send_keepalive(command_con);
+
+    assert(tick_event == this);
+    tick_event = new C_Tick(ctx);
+    ctx->timer.add_event_after(tick_interval, tick_event);
+  }
 };
 
-void send_observe_requests(CephToolCtx *ctx)
-{
-  dout(1) << "send_observe_requests " << dendl;
-
-  for (int i=0; i<PAXOS_NUM; i++) {
-    MMonObserve *m = new MMonObserve(ctx->mc.monmap.fsid, i, map_ver[i]);
-    dout(1) << "mon" << " <- observe " << get_paxos_name(i) << dendl;
-    ctx->mc.send_mon_message(m);
-  }
-
-  registered.clear();
-  float seconds = g_conf->paxos_observer_timeout/2;
-  dout(1) << " refresh after " << seconds << " with same mon" << dendl;
-  ctx->timer.add_event_after(seconds, new C_ObserverRefresh(false, ctx));
-}
-
-static void handle_ack(CephToolCtx *ctx, MMonCommandAck *ack)
-{
-  ctx->lock.Lock();
-  reply = true;
-  reply_from = ack->get_source_inst();
-  reply_rs = ack->rs;
-  reply_rc = ack->r;
-  reply_bl = ack->get_data();
-  cmd_cond.Signal();
-  if (resend_event) {
-    ctx->timer.cancel_event(resend_event);
-    resend_event = 0;
-  }
-  ctx->lock.Unlock();
-  ack->put();
-}
-
-static void handle_ack(CephToolCtx *ctx, MCommandReply *ack)
-{
-  ctx->lock.Lock();
-  if (ack->get_tid() == pending_tid) {
-    reply = true;
-    reply_from = ack->get_source_inst();
-    reply_rs = ack->rs;
-    reply_rc = ack->r;
-    reply_bl = ack->get_data();
-    cmd_cond.Signal();
-    if (resend_event) {
-      ctx->timer.cancel_event(resend_event);
-      resend_event = 0;
-    }
-  }
-  ctx->lock.Unlock();
-  ack->put();
-}
-
 static void send_command(CephToolCtx *ctx)
 {
   if (!pending_tell && !pending_tell_pgid) {
@@ -322,6 +122,15 @@ static void send_command(CephToolCtx *ctx)
     return;
   }
 
+  if (pending_tell_pgid || pending_target.get_type() == (int)entity_name_t::TYPE_OSD) {
+    if (!osdmap) {
+      ctx->mc.sub_want("osdmap", 0, CEPH_SUBSCRIBE_ONETIME);
+      ctx->mc.renew_subs();
+      cmd_waiting_for_osdmap = true;
+      return;
+    }
+  }
+
   if (!ctx->concise)
     *ctx->log << ceph_clock_now(g_ceph_context) << " " << pending_target << " <- " << pending_cmd << std::endl;
 
@@ -330,15 +139,6 @@ static void send_command(CephToolCtx *ctx)
   m->set_data(pending_bl);
   m->set_tid(++pending_tid);
 
-  if (pending_tell_pgid || pending_target.get_type() == (int)entity_name_t::TYPE_OSD) {
-    if (!osdmap) {
-      ctx->mc.sub_want("osdmap", 0, CEPH_SUBSCRIBE_ONETIME);
-      ctx->mc.renew_subs();
-      while (!osdmap)
-	cmd_cond.Wait(ctx->lock);
-    }
-  }
-
   if (pending_tell_pgid) {
     // pick target osd
     vector<int> osds;
@@ -376,7 +176,13 @@ static void send_command(CephToolCtx *ctx)
       if (!ctx->concise)
 	*ctx->log << ceph_clock_now(g_ceph_context) << " " << pending_target << " <- " << pending_cmd << std::endl;
 
-      messenger->send_message(m, osdmap->get_inst(n));
+      command_con = messenger->get_connection(osdmap->get_inst(n));
+      messenger->send_message(m, command_con);
+
+      if (tick_event)
+	ctx->timer.cancel_event(tick_event);
+      tick_event = new C_Tick(ctx);
+      ctx->timer.add_event_after(tick_interval, tick_event);
     }
     return;
   }
@@ -385,65 +191,53 @@ static void send_command(CephToolCtx *ctx)
   reply = true;
 }
 
-class Admin : public Dispatcher {
-public:
-  Admin(CephToolCtx *ctx_)
-    : Dispatcher(g_ceph_context),
-      ctx(ctx_)
-  {
+static void handle_osd_map(CephToolCtx *ctx, MOSDMap *m)
+{
+  epoch_t e = m->get_first();
+  assert(m->maps.count(e));
+  ctx->lock.Lock();
+  delete osdmap;
+  osdmap = new OSDMap;
+  osdmap->decode(m->maps[e]);
+  if (cmd_waiting_for_osdmap) {
+    cmd_waiting_for_osdmap = false;
+    send_command(ctx);
   }
+  ctx->lock.Unlock();
+  m->put();
+}
 
-  bool ms_dispatch(Message *m) {
-    switch (m->get_type()) {
-    case MSG_MON_COMMAND_ACK:
-      handle_ack(ctx, (MMonCommandAck*)m);
-      break;
-    case MSG_COMMAND_REPLY:
-      handle_ack(ctx, (MCommandReply*)m);
-      break;
-    case MSG_MON_OBSERVE_NOTIFY:
-      handle_notify(ctx, (MMonObserveNotify *)m);
-      break;
-    case MSG_MON_OBSERVE:
-      handle_observe(ctx, (MMonObserve *)m);
-      break;
-    case CEPH_MSG_MON_MAP:
-      m->put();
-      break;
-    case CEPH_MSG_OSD_MAP:
-      handle_osd_map(ctx, (MOSDMap *)m);
-      break;
-    default:
-      return false;
-    }
-    return true;
-  }
+static void handle_ack(CephToolCtx *ctx, MMonCommandAck *ack)
+{
+  ctx->lock.Lock();
+  reply = true;
+  reply_from = ack->get_source_inst();
+  reply_rs = ack->rs;
+  reply_rc = ack->r;
+  reply_bl = ack->get_data();
+  cmd_cond.Signal();
+  ctx->lock.Unlock();
+  ack->put();
+}
 
-  void ms_handle_connect(Connection *con) {
-    if (con->get_peer_type() == CEPH_ENTITY_TYPE_MON) {
-      ctx->lock.Lock();
-      if (ceph_tool_mode != CEPH_TOOL_MODE_CLI_INPUT) {
-	send_observe_requests(ctx);
-      }
-      if (pending_cmd.size())
-	send_command(ctx);
-      ctx->lock.Unlock();
+static void handle_ack(CephToolCtx *ctx, MCommandReply *ack)
+{
+  ctx->lock.Lock();
+  if (ack->get_tid() == pending_tid) {
+    reply = true;
+    reply_from = ack->get_source_inst();
+    reply_rs = ack->rs;
+    reply_rc = ack->r;
+    reply_bl = ack->get_data();
+    cmd_cond.Signal();
+    if (tick_event) {
+      ctx->timer.cancel_event(tick_event);
+      tick_event = 0;
     }
   }
-  bool ms_handle_reset(Connection *con) { return false; }
-  void ms_handle_remote_reset(Connection *con) {}
-
-  bool ms_get_authorizer(int dest_type, AuthAuthorizer **authorizer, bool force_new)
-  {
-    if (dest_type == CEPH_ENTITY_TYPE_MON)
-      return true;
-    *authorizer = ctx->mc.auth->build_authorizer(dest_type);
-    return true;
-  }
-
-private:
-  CephToolCtx *ctx;
-};
+  ctx->lock.Unlock();
+  ack->put();
+}
 
 int do_command(CephToolCtx *ctx,
 	       vector<string>& cmd, bufferlist& bl, bufferlist& rbl)
@@ -497,6 +291,17 @@ int do_command(CephToolCtx *ctx,
   return reply_rc;
 }
 
+void do_status(CephToolCtx *ctx, bool shutdown) {
+  vector<string> cmd;
+  cmd.push_back("status");
+  bufferlist bl;
+
+  do_command(ctx, cmd, bl, bl);
+
+  if (shutdown)
+    messenger->shutdown();
+}
+
 static const char *cli_prompt(EditLine *e)
 {
   return "ceph> ";
@@ -654,11 +459,14 @@ CephToolCtx* ceph_tool_common_init(ceph_tool_mode_t mode, bool concise)
   // start up network
   messenger = new SimpleMessenger(g_ceph_context, entity_name_t::CLIENT(), "client",
                                   getpid());
+  messenger->set_default_policy(Messenger::Policy::lossy_client(0, 0));
   messenger->start();
+
+  ctx->mc.set_messenger(messenger);
+
   ctx->dispatcher = new Admin(ctx.get());
   messenger->add_dispatcher_head(ctx->dispatcher);
 
-  ctx->mc.set_messenger(messenger);
   int r = ctx->mc.init();
   if (r < 0)
     return NULL;
@@ -705,3 +513,85 @@ int ceph_tool_common_shutdown(CephToolCtx *ctx)
   
   return 0;
 }
+
+void Subscriptions::handle_log(MLog *m) 
+{
+  dout(10) << __func__ << " received log msg ver " << m->version << dendl;
+
+  if (last_known_version >= m->version) {
+    dout(10) << __func__ 
+	    << " we have already received ver " << m->version 
+	    << " (highest received: " << last_known_version << ") " << dendl;
+    return;
+  }
+  last_known_version = m->version;
+
+  std::deque<LogEntry>::iterator it = m->entries.begin();
+  for (; it != m->entries.end(); it++) {
+    LogEntry e = *it;
+    cout << e.stamp << " " << e.who.name 
+	 << " " << e.type << " " << e.msg << std::endl;
+  }
+ 
+  version_t v = last_known_version+1;
+  dout(10) << __func__ << " wanting " << name << " ver " << v << dendl;
+  ctx->mc.sub_want(name, v, 0);
+}
+
+bool Admin::ms_dispatch(Message *m) {
+  switch (m->get_type()) {
+  case MSG_MON_COMMAND_ACK:
+    handle_ack(ctx, (MMonCommandAck*)m);
+    break;
+  case MSG_COMMAND_REPLY:
+    handle_ack(ctx, (MCommandReply*)m);
+    break;
+  case CEPH_MSG_MON_MAP:
+    m->put();
+    break;
+  case CEPH_MSG_OSD_MAP:
+    handle_osd_map(ctx, (MOSDMap *)m);
+    break;
+  case MSG_LOG:
+   {
+    MLog *mlog = (MLog*) m;
+    subs.handle_log(mlog);
+    break;
+   }
+  default:
+    return false;
+  }
+  return true;
+}
+
+void Admin::ms_handle_connect(Connection *con) {
+  if (con->get_peer_type() == CEPH_ENTITY_TYPE_MON) {
+    ctx->lock.Lock();
+    if (pending_cmd.size())
+      send_command(ctx);
+    ctx->lock.Unlock();
+  }
+}
+
+bool Admin::ms_handle_reset(Connection *con)
+{
+  Mutex::Locker l(ctx->lock);
+  if (con == command_con) {
+    command_con->put();
+    command_con = NULL;
+    if (pending_cmd.size())
+      send_command(ctx);
+    return true;
+  }
+  return false;
+}
+
+bool Admin::ms_get_authorizer(int dest_type, AuthAuthorizer **authorizer, 
+			      bool force_new)
+{
+  if (dest_type == CEPH_ENTITY_TYPE_MON)
+    return true;
+  *authorizer = ctx->mc.auth->build_authorizer(dest_type);
+  return true;
+}
+
